import pytest

from ufl import Coefficient, FunctionSpace, Index, Mesh, as_tensor, interval, sqrt, triangle
from ufl.algorithms.renumbering import renumber_indices
from ufl.compound_expressions import cross_expr, determinant_expr, inverse_expr
from ufl.finiteelement import FiniteElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1


@pytest.fixture
def A0(request):
    return Coefficient(
        FunctionSpace(
            Mesh(FiniteElement("Lagrange", interval, 1, (1,), identity_pullback, H1)),
            FiniteElement("Lagrange", interval, 1, (), identity_pullback, H1),
        )
    )


@pytest.fixture
def A1(request):
    return Coefficient(
        FunctionSpace(
            Mesh(FiniteElement("Lagrange", interval, 1, (1,), identity_pullback, H1)),
            FiniteElement("Lagrange", interval, 1, (1, 1), identity_pullback, H1),
        )
    )


@pytest.fixture
def A2(request):
    return Coefficient(
        FunctionSpace(
            Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1)),
            FiniteElement("Lagrange", triangle, 1, (2, 2), identity_pullback, H1),
        )
    )


@pytest.fixture
def A3(request):
    return Coefficient(
        FunctionSpace(
            Mesh(FiniteElement("Lagrange", triangle, 1, (3,), identity_pullback, H1)),
            FiniteElement("Lagrange", triangle, 1, (3, 3), identity_pullback, H1),
        )
    )


@pytest.fixture
def A21(request):
    return Coefficient(
        FunctionSpace(
            Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1)),
            FiniteElement("Lagrange", triangle, 1, (2, 1), identity_pullback, H1),
        )
    )


@pytest.fixture
def A31(request):
    return Coefficient(
        FunctionSpace(
            Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1)),
            FiniteElement("Lagrange", triangle, 1, (3, 1), identity_pullback, H1),
        )
    )


@pytest.fixture
def A32(request):
    return Coefficient(
        FunctionSpace(
            Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1)),
            FiniteElement("Lagrange", triangle, 1, (3, 2), identity_pullback, H1),
        )
    )


def test_determinant0(A0):
    assert determinant_expr(A0) == A0


def test_determinant1(A1):
    assert determinant_expr(A1) == A1[0, 0]


def test_determinant2(A2):
    assert determinant_expr(A2) == A2[0, 0] * A2[1, 1] - A2[0, 1] * A2[1, 0]


def test_determinant3(A3):
    assert determinant_expr(A3) == (
        A3[0, 0] * (A3[1, 1] * A3[2, 2] - A3[1, 2] * A3[2, 1])
        + (A3[1, 0] * A3[2, 2] - A3[1, 2] * A3[2, 0]) * (-A3[0, 1])
        + A3[0, 2] * (A3[1, 0] * A3[2, 1] - A3[1, 1] * A3[2, 0])
    )


def test_pseudo_determinant21(A21):
    i = Index()
    assert renumber_indices(determinant_expr(A21)) == renumber_indices(sqrt(A21[i, 0] * A21[i, 0]))


def test_pseudo_determinant31(A31):
    i = Index()
    assert renumber_indices(determinant_expr(A31)) == renumber_indices(
        sqrt((A31[i, 0] * A31[i, 0]))
    )


def test_pseudo_determinant32(A32):
    i = Index()
    c = cross_expr(A32[:, 0], A32[:, 1])
    assert renumber_indices(determinant_expr(A32)) == renumber_indices(sqrt(c[i] * c[i]))


def test_inverse0(A0):
    expected = 1.0 / A0  # stays scalar
    assert inverse_expr(A0) == renumber_indices(expected)


def test_inverse1(A1):
    expected = as_tensor(((1.0 / A1[0, 0],),))  # reshaped into 1x1 tensor
    assert inverse_expr(A1) == renumber_indices(expected)


def xtest_inverse2(A2):
    expected = "TODO"
    assert inverse_expr(A2) == renumber_indices(expected)


def xtest_inverse3(A3):
    expected = "TODO"
    assert inverse_expr(A3) == renumber_indices(expected)


def xtest_pseudo_inverse21(A21):
    expected = "TODO"
    assert renumber_indices(inverse_expr(A21)) == renumber_indices(expected)


def xtest_pseudo_inverse31(A31):
    expected = "TODO"
    assert renumber_indices(inverse_expr(A31)) == renumber_indices(expected)


def xtest_pseudo_inverse32(A32):
    expected = "TODO"
    assert renumber_indices(inverse_expr(A32)) == renumber_indices(expected)



"""Transformer.

This module defines the Transformer base class and some
basic specializations to further base other algorithms upon,
as well as some utilities for easier application of such
algorithms.
"""
# Copyright (C) 2008-2016 Martin Sandve Alnæs and Anders Logg
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Anders Logg, 2009-2010

import inspect

from ufl.algorithms.map_integrands import map_integrands
from ufl.classes import Variable, all_ufl_classes
from ufl.core.ufl_type import UFLType


def is_post_handler(function):
    """Check if function is a handler that expects transformed children as input."""
    insp = inspect.getfullargspec(function)
    num_args = len(insp[0]) + int(insp[1] is not None)
    visit_children_first = num_args > 2
    return visit_children_first


class Transformer(object):
    """Transformer.

    Base class for a visitor-like algorithm design pattern used to
    transform expression trees from one representation to another.
    """

    _handlers_cache = {}

    def __init__(self, variable_cache=None):
        """Initialise."""
        if variable_cache is None:
            variable_cache = {}
        self._variable_cache = variable_cache

        # Analyse class properties and cache handler data the
        # first time this is run for a particular class
        cache_data = Transformer._handlers_cache.get(type(self))
        if not cache_data:
            cache_data = [None] * len(all_ufl_classes)
            # For all UFL classes
            for classobject in all_ufl_classes:
                # Iterate over the inheritance chain
                # (NB! This assumes that all UFL classes inherits a single
                # Expr subclass and that this is the first superclass!)
                for c in classobject.mro():
                    # Register classobject with handler for the first
                    # encountered superclass
                    try:
                        handler_name = c._ufl_handler_name_
                    except AttributeError as attribute_error:
                        if type(classobject) is not UFLType:
                            raise attribute_error
                        # Default handler name for UFL types
                        handler_name = UFLType._ufl_handler_name_
                    function = getattr(self, handler_name, None)
                    if function:
                        cache_data[classobject._ufl_typecode_] = (
                            handler_name,
                            is_post_handler(function),
                        )
                        break
            Transformer._handlers_cache[type(self)] = cache_data

        # Build handler list for this particular class (get functions
        # bound to self)
        self._handlers = [(getattr(self, name), post) for (name, post) in cache_data]
        # Keep a stack of objects visit is called on, to ease
        # backtracking
        self._visit_stack = []

    def print_visit_stack(self):
        """Print visit stack."""
        print("/" * 80)
        print("Visit stack in Transformer:")

        def sstr(s):
            """Format."""
            ss = str(type(s)) + " ; "
            n = 160 - len(ss)
            return ss + str(s)[:n]

        print("\n".join(sstr(s) for s in self._visit_stack))
        print("\\" * 80)

    def visit(self, o):
        """Visit."""
        # Update stack
        self._visit_stack.append(o)

        # Get handler for the UFL class of o (type(o) may be an
        # external subclass of the actual UFL class)
        h, visit_children_first = self._handlers[o._ufl_typecode_]

        # Is this a handler that expects transformed children as
        # input?
        if visit_children_first:
            # Yes, visit all children first and then call h.
            r = h(o, *[self.visit(op) for op in o.ufl_operands])
        else:
            # No, this is a handler that handles its own children
            # (arguments self and o, where self is already bound)
            r = h(o)

        # Update stack and return
        self._visit_stack.pop()
        return r

    def undefined(self, o):
        """Trigger error."""
        raise ValueError(f"No handler defined for {o._ufl_class_.__name__}.")

    def reuse(self, o):
        """Reuse Expr (ignore children)."""
        return o

    def reuse_if_untouched(self, o, *ops):
        """Reuse object if operands are the same objects.

        Use in your own subclass by setting e.g. `expr = MultiFunction.reuse_if_untouched`
        as a default rule.
        """
        if all(a is b for a, b in zip(o.ufl_operands, ops)):
            return o
        else:
            return o._ufl_expr_reconstruct_(*ops)

    # It's just so slow to compare all operands, avoiding it now
    reuse_if_possible = reuse_if_untouched

    def always_reconstruct(self, o, *operands):
        """Reconstruct expr."""
        return o._ufl_expr_reconstruct_(*operands)

    # Set default behaviour for any UFLType
    ufl_type = undefined

    # Set default behaviour for any Terminal
    terminal = reuse

    def reuse_variable(self, o):
        """Reuse variable."""
        # Check variable cache to reuse previously transformed
        # variable if possible
        e, l = o.ufl_operands  # noqa: E741
        v = self._variable_cache.get(l)
        if v is not None:
            return v

        # Visit the expression our variable represents
        e2 = self.visit(e)

        # If the expression is the same, reuse Variable object
        if e == e2:
            v = o
        else:
            # Recreate Variable (with same label)
            v = Variable(e2, l)

        # Cache variable
        self._variable_cache[l] = v
        return v

    def reconstruct_variable(self, o):
        """Reconstruct variable."""
        # Check variable cache to reuse previously transformed
        # variable if possible
        e, l = o.ufl_operands  # noqa: E741
        v = self._variable_cache.get(l)
        if v is not None:
            return v

        # Visit the expression our variable represents
        e2 = self.visit(e)

        # Always reconstruct Variable (with same label)
        v = Variable(e2, l)
        self._variable_cache[l] = v
        return v


class ReuseTransformer(Transformer):
    """Reuse transformer."""

    def __init__(self, variable_cache=None):
        """Initialise."""
        Transformer.__init__(self, variable_cache)

    # Set default behaviour for any Expr
    expr = Transformer.reuse_if_untouched

    # Set default behaviour for any Terminal
    terminal = Transformer.reuse

    # Set default behaviour for Variable
    variable = Transformer.reuse_variable


class CopyTransformer(Transformer):
    """Copy transformer."""

    def __init__(self, variable_cache=None):
        """Initialise."""
        Transformer.__init__(self, variable_cache)

    # Set default behaviour for any Expr
    expr = Transformer.always_reconstruct

    # Set default behaviour for any Terminal
    terminal = Transformer.reuse

    # Set default behaviour for Variable
    variable = Transformer.reconstruct_variable


class VariableStripper(ReuseTransformer):
    """Variable stripper."""

    def __init__(self):
        """Initialise."""
        ReuseTransformer.__init__(self)

    def variable(self, o):
        """Visit a variable."""
        return self.visit(o.ufl_operands[0])


def apply_transformer(e, transformer, integral_type=None):
    """Apply transforms.

    Apply transformer.visit(expression) to each integrand expression in
    form, or to form if it is an Expr.
    """
    return map_integrands(lambda expr: transformer.visit(expr), e, integral_type)


def strip_variables(e):
    """Replace all Variable instances with the expression they represent."""
    return apply_transformer(e, VariableStripper())



"""Various string formatting utilities."""
# Copyright (C) 2008-2016 Martin Sandve Alnæs and Anders Logg
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later


def camel2underscore(name):
    """Convert a CamelCaps string to underscore_syntax."""
    letters = []
    lastlower = False
    for i in name:
        thislower = i.islower() or i.isdigit()
        if not thislower:
            # Don't insert _ between multiple upper case letters
            if lastlower:
                letters.append("_")
            i = i.lower()
        lastlower = thislower
        letters.append(i)
    return "".join(letters)


def lstr(a):
    """Pretty-print list or tuple, invoking str() on items instead of repr() like str() does."""
    if isinstance(a, list):
        return "[" + ", ".join(lstr(item) for item in a) + "]"
    elif isinstance(a, tuple):
        return "(" + ", ".join(lstr(item) for item in a) + ")"
    return str(a)


def tstr(t, colsize=80):
    """Pretty-print list of tuples of key-value pairs."""
    if not t:
        return ""

    # Compute maximum key length
    keylen = max(len(str(item[0])) for item in t)

    # Key-length cannot be larger than colsize
    if keylen > colsize:
        return str(t)

    # Pretty-print table
    s = ""
    for key, value in t:
        key = str(key)
        if isinstance(value, str):
            value = "'%s'" % value
        else:
            value = str(value)
        s += key + ":" + " " * (keylen - len(key) + 1)
        space = ""
        while len(value) > 0:
            end = min(len(value), colsize - keylen)
            s += space + value[:end] + "\n"
            value = value[end:]
            space = " " * (keylen + 2)
    return s


def istr(o):
    """Format object as string, inserting ? for None."""
    if o is None:
        return "?"
    else:
        return str(o)


def estr(elements):
    """Format list of elements for printing."""
    return ", ".join(f"{e}" for e in elements)


def _indent_string(n):
    """Return indentation string."""
    return "    " * n


def _tree_format_expression(expression, indentation, parentheses):
    """Tree format expression."""
    ind = _indent_string(indentation)
    if expression._ufl_is_terminal_:
        s = "%s%s" % (ind, repr(expression))
    else:
        sops = [
            _tree_format_expression(o, indentation + 1, parentheses)
            for o in expression.ufl_operands
        ]
        s = "%s%s\n" % (ind, expression._ufl_class_.__name__)
        if parentheses and len(sops) > 1:
            s += "%s(\n" % (ind,)
        s += "\n".join(sops)
        if parentheses and len(sops) > 1:
            s += "\n%s)" % (ind,)
    return s


def tree_format(expression, indentation=0, parentheses=True):
    """Tree format."""
    from ufl.core.expr import Expr
    from ufl.form import Form
    from ufl.integral import Integral

    s = ""

    if isinstance(expression, Form):
        form = expression
        integrals = form.integrals()
        integral_types = sorted(set(itg.integral_type() for itg in integrals))
        itgs = []
        for integral_type in integral_types:
            itgs += list(form.integrals_by_type(integral_type))

        ind = _indent_string(indentation)
        s += ind + "Form:\n"
        s += "\n".join(tree_format(itg, indentation + 1, parentheses) for itg in itgs)

    elif isinstance(expression, Integral):
        ind = _indent_string(indentation)
        s += ind + "Integral:\n"
        ind = _indent_string(indentation + 1)
        s += ind + "integral type: %s\n" % expression.integral_type()
        s += ind + "subdomain id: %s\n" % expression.subdomain_id()
        s += ind + "integrand:\n"
        s += tree_format(expression._integrand, indentation + 2, parentheses)

    elif isinstance(expression, Expr):
        s += _tree_format_expression(expression, indentation, parentheses)

    else:
        raise ValueError(f"Invalid object type {type(expression)}")

    return s



__authors__ = "Martin Sandve Alnæs"
__date__ = "2009-03-19 -- 2012-03-20"

# Modified by Anders Logg, 2008
# Modified by Garth N. Wells, 2009

import math

import pytest

from ufl import (
    Coefficient,
    FunctionSpace,
    Identity,
    Mesh,
    as_tensor,
    cos,
    det,
    div,
    dot,
    dx,
    exp,
    grad,
    i,
    inner,
    j,
    k,
    l,
    ln,
    nabla_div,
    nabla_grad,
    outer,
    sin,
    triangle,
)
from ufl.algorithms import compute_form_data, expand_derivatives, expand_indices
from ufl.algorithms.renumbering import renumber_indices
from ufl.finiteelement import FiniteElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1

# TODO: Test expand_indices2 throuroughly for correctness, then efficiency:
# expand_indices, expand_indices2 = expand_indices2, expand_indices


class Fixture:
    def __init__(self):
        cell = triangle
        element = FiniteElement("Lagrange", cell, 1, (), identity_pullback, H1)
        velement = FiniteElement("Lagrange", cell, 1, (2,), identity_pullback, H1)
        telement = FiniteElement("Lagrange", cell, 1, (2, 2), identity_pullback, H1)
        domain = Mesh(FiniteElement("Lagrange", cell, 1, (2,), identity_pullback, H1))
        space = FunctionSpace(domain, element)
        vspace = FunctionSpace(domain, velement)
        tspace = FunctionSpace(domain, telement)
        self.sf = Coefficient(space)
        self.sf2 = Coefficient(space)
        self.vf = Coefficient(vspace)
        self.tf = Coefficient(tspace)

        # Note: the derivatives of these functions make no sense, but
        #       their unique constant values are used for validation.

        def SF(x, derivatives=()):
            # first order derivatives
            if derivatives == (0,):
                return 0.30
            elif derivatives == (1,):
                return 0.31
            # second order derivatives
            elif derivatives == (0, 0):
                return 0
            elif derivatives in ((1, 0), (0, 1)):
                return 0
            elif derivatives == (1, 1):
                return 0
            # function value
            assert derivatives == ()
            return 3

        def SF2(x, derivatives=()):
            # first order derivatives
            if derivatives == (0,):
                return 0.30
            elif derivatives == (1,):
                return 0.31
            # second order derivatives
            elif derivatives == (0, 0):
                return 3.300
            elif derivatives in ((1, 0), (0, 1)):
                return 3.310
            elif derivatives == (1, 1):
                return 3.311
            # function value
            assert derivatives == ()
            return 3

        def VF(x, derivatives=()):
            # first order derivatives
            if derivatives == (0,):
                return (0.50, 0.70)
            elif derivatives == (1,):
                return (0.51, 0.71)
            # second order derivatives
            elif derivatives == (0, 0):
                return (0.20, 0.21)
            elif derivatives in ((1, 0), (0, 1)):
                return (0.30, 0.31)
            elif derivatives == (1, 1):
                return (0.40, 0.41)
            # function value
            assert derivatives == ()
            return (5, 7)

        def TF(x, derivatives=()):
            # first order derivatives
            if derivatives == (0,):
                return ((1.10, 1.30), (1.70, 1.90))
            elif derivatives == (1,):
                return ((1.11, 1.31), (1.71, 1.91))
            # second order derivatives
            elif derivatives == (0, 0):
                return ((10.00, 10.01), (10.10, 10.11))
            elif derivatives in ((1, 0), (0, 1)):
                return ((12.00, 12.01), (12.10, 12.11))
            elif derivatives == (1, 1):
                return ((11.00, 11.01), (11.10, 11.11))
            # function value
            assert derivatives == ()
            return ((11, 13), (17, 19))

        self.x = (1.23, 3.14)
        self.mapping = {self.sf: SF, self.sf2: SF2, self.vf: VF, self.tf: TF}

    def compare(self, f, value):
        debug = 0
        if debug:
            print(("f", f))
        g = expand_derivatives(f)
        if debug:
            print(("g", g))
        gv = g(self.x, self.mapping)
        assert abs(gv - value) < 1e-7

        g = expand_indices(g)
        if debug:
            print(("g", g))
        gv = g(self.x, self.mapping)
        assert abs(gv - value) < 1e-7

        g = renumber_indices(g)
        if debug:
            print(("g", g))
        gv = g(self.x, self.mapping)
        assert abs(gv - value) < 1e-7


@pytest.fixture(scope="module")
def fixt():
    # Workaround for quick pytest transition
    return Fixture()


def test_basic_expand_indices(self, fixt):
    sf = fixt.sf
    vf = fixt.vf
    tf = fixt.tf
    compare = fixt.compare

    # Simple expressions with no indices or derivatives to expand
    compare(sf, 3)
    compare(sf + 1, 4)
    compare(sf - 2.5, 0.5)
    compare(sf / 2, 1.5)
    compare(sf / 0.5, 6)
    compare(sf**2, 9)
    compare(sf**0.5, 3**0.5)
    compare(sf**3, 27)
    compare(0.5**sf, 0.5**3)
    compare(sf * (sf / 6), 1.5)
    compare(sin(sf), math.sin(3))
    compare(cos(sf), math.cos(3))
    compare(exp(sf), math.exp(3))
    compare(ln(sf), math.log(3))

    # Simple indexing
    compare(vf[0], 5)
    compare(vf[0] + 1, 6)
    compare(vf[0] - 2.5, 2.5)
    compare(vf[0] / 2, 2.5)
    compare(vf[0] / 0.5, 10)
    compare(vf[0] ** 2, 25)
    compare(vf[0] ** 0.5, 5**0.5)
    compare(vf[0] ** 3, 125)
    compare(0.5 ** vf[0], 0.5**5)
    compare(vf[0] * (vf[0] / 6), 5 * (5.0 / 6))
    compare(sin(vf[0]), math.sin(5))
    compare(cos(vf[0]), math.cos(5))
    compare(exp(vf[0]), math.exp(5))
    compare(ln(vf[0]), math.log(5))

    # Double indexing
    compare(tf[1, 1], 19)
    compare(tf[1, 1] + 1, 20)
    compare(tf[1, 1] - 2.5, 16.5)
    compare(tf[1, 1] / 2, 9.5)
    compare(tf[1, 1] / 0.5, 38)
    compare(tf[1, 1] ** 2, 19**2)
    compare(tf[1, 1] ** 0.5, 19**0.5)
    compare(tf[1, 1] ** 3, 19**3)
    compare(0.5 ** tf[1, 1], 0.5**19)
    compare(tf[1, 1] * (tf[1, 1] / 6), 19 * (19.0 / 6))
    compare(sin(tf[1, 1]), math.sin(19))
    compare(cos(tf[1, 1]), math.cos(19))
    compare(exp(tf[1, 1]), math.exp(19))
    compare(ln(tf[1, 1]), math.log(19))


def test_expand_indices_index_sum(self, fixt):
    vf = fixt.vf
    tf = fixt.tf
    compare = fixt.compare

    # Basic index sums
    compare(vf[i] * vf[i], 5 * 5 + 7 * 7)
    compare(vf[j] * tf[i, j] * vf[i], 5 * 5 * 11 + 5 * 7 * 13 + 5 * 7 * 17 + 7 * 7 * 19)
    compare(vf[j] * tf.T[j, i] * vf[i], 5 * 5 * 11 + 5 * 7 * 13 + 5 * 7 * 17 + 7 * 7 * 19)
    compare(tf[i, i], 11 + 19)
    compare(
        tf[i, j] * (tf[j, i] + outer(vf, vf)[i, j]),
        (5 * 5 + 11) * 11 + (7 * 5 + 17) * 13 + (7 * 5 + 13) * 17 + (7 * 7 + 19) * 19,
    )
    compare(as_tensor(as_tensor(tf[i, j], (i, j))[k, l], (l, k))[i, i], 11 + 19)


def test_expand_indices_derivatives(self, fixt):
    sf = fixt.sf
    vf = fixt.vf
    compare = fixt.compare

    # Basic derivatives
    compare(sf.dx(0), 0.3)
    compare(sf.dx(1), 0.31)
    compare(sf.dx(i) * vf[i], 0.30 * 5 + 0.31 * 7)
    compare(vf[j].dx(i) * vf[i].dx(j), 0.50 * 0.50 + 0.51 * 0.70 + 0.70 * 0.51 + 0.71 * 0.71)


def test_expand_indices_hyperelasticity(self, fixt):
    vf = fixt.vf
    compare = fixt.compare

    # Deformation gradient
    ident = Identity(2)
    u = vf
    F = ident + grad(u)
    # F = (1 + vf[0].dx(0), vf[0].dx(1), vf[1].dx(0), 1 + vf[1].dx(1))
    # F = (1 + 0.50,        0.51,        0.70,        1 + 0.71)
    F00 = 1 + 0.50
    F01 = 0.51
    F10 = 0.70
    F11 = 1 + 0.71
    compare(F[0, 0], F00)
    compare(F[0, 1], F01)
    compare(F[1, 0], F10)
    compare(F[1, 1], F11)

    J = det(F)
    compare(J, (1 + 0.50) * (1 + 0.71) - 0.70 * 0.51)

    # Strain tensors
    C = F.T * F
    # Cij = sum_k Fki Fkj
    C00 = F00 * F00 + F10 * F10
    C01 = F00 * F01 + F10 * F11
    C10 = F01 * F00 + F11 * F10
    C11 = F01 * F01 + F11 * F11
    compare(C[0, 0], C00)
    compare(C[0, 1], C01)
    compare(C[1, 0], C10)
    compare(C[1, 1], C11)

    E = (C - ident) / 2
    E00 = (C00 - 1) / 2
    E01 = (C01) / 2
    E10 = (C10) / 2
    E11 = (C11 - 1) / 2
    compare(E[0, 0], E00)
    compare(E[0, 1], E01)
    compare(E[1, 0], E10)
    compare(E[1, 1], E11)

    # Strain energy
    Q = inner(E, E)
    Qvalue = E00**2 + E01**2 + E10**2 + E11**2
    compare(Q, Qvalue)

    K = 0.5
    psi = (K / 2) * exp(Q)
    compare(psi, 0.25 * math.exp(Qvalue))


def test_expand_indices_div_grad(self, fixt):
    sf = fixt.sf
    sf2 = fixt.sf2
    vf = fixt.vf
    tf = fixt.tf
    compare = fixt.compare

    a = div(grad(sf))
    compare(a, 0)

    a = div(grad(sf2))
    compare(a, 3.300 + 3.311)

    if 0:
        Dvf = grad(vf)
        Lvf = div(Dvf)
        Lvf2 = dot(Lvf, Lvf)
        pp = compute_form_data(Lvf2 * dx).preprocessed_form.integrals()[0].integrand()
        print(("vf", vf.ufl_shape, str(vf)))
        print(("Dvf", Dvf.ufl_shape, str(Dvf)))
        print(("Lvf", Lvf.ufl_shape, str(Lvf)))
        print(("Lvf2", Lvf2.ufl_shape, str(Lvf2)))
        print(("pp", pp.ufl_shape, str(pp)))

    a = div(grad(vf))
    compare(dot(a, a), (0.20 + 0.40) ** 2 + (0.21 + 0.41) ** 2)

    a = div(grad(tf))
    compare(
        inner(a, a),
        (10.00 + 11.00) ** 2 + (10.01 + 11.01) ** 2 + (10.10 + 11.10) ** 2 + (10.11 + 11.11) ** 2,
    )


def test_expand_indices_nabla_div_grad(self, fixt):
    sf = fixt.sf
    sf2 = fixt.sf2
    vf = fixt.vf
    tf = fixt.tf
    compare = fixt.compare

    a = nabla_div(nabla_grad(sf))
    compare(a, 0)

    a = nabla_div(nabla_grad(sf2))
    compare(a, 3.300 + 3.311)

    a = nabla_div(nabla_grad(vf))
    compare(dot(a, a), (0.20 + 0.40) ** 2 + (0.21 + 0.41) ** 2)

    a = nabla_div(nabla_grad(tf))
    compare(
        inner(a, a),
        (10.00 + 11.00) ** 2 + (10.01 + 11.01) ** 2 + (10.10 + 11.10) ** 2 + (10.11 + 11.11) ** 2,
    )



"""Restriction operations."""
# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

from ufl.core.operator import Operator
from ufl.core.ufl_type import ufl_type
from ufl.precedence import parstr

# --- Restriction operators ---


@ufl_type(
    is_abstract=True,
    num_ops=1,
    inherit_shape_from_operand=0,
    inherit_indices_from_operand=0,
    is_restriction=True,
)
class Restricted(Operator):
    """Restriction."""

    __slots__ = ()

    # TODO: Add __new__ operator here, e.g. restricted(literal) == literal

    def __init__(self, f):
        """Initialise."""
        Operator.__init__(self, (f,))

    def side(self):
        """Get the side."""
        return self._side

    def evaluate(self, x, mapping, component, index_values):
        """Evaluate."""
        return self.ufl_operands[0].evaluate(x, mapping, component, index_values)

    def __str__(self):
        """Format as a string."""
        return f"{parstr(self.ufl_operands[0], self)}({self._side})"


@ufl_type(is_terminal_modifier=True)
class PositiveRestricted(Restricted):
    """Positive restriction."""

    __slots__ = ()
    _side = "+"


@ufl_type(is_terminal_modifier=True)
class NegativeRestricted(Restricted):
    """Negative restriction."""

    __slots__ = ()
    _side = "-"



"""Various utility data structures."""

# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later


class Stack(list):
    """A stack datastructure."""

    def __init__(self, *args):
        """Initialise."""
        list.__init__(self, *args)

    def push(self, v):
        """Push."""
        list.append(self, v)

    def peek(self):
        """Peek."""
        return self[-1]


class StackDict(dict):
    """A dictionary type.

    A dict that can be changed incrementally with 'd.push(k,v)' and have
    changes rolled back with 'k,v = d.pop()'.
    """

    def __init__(self, *args, **kwargs):
        """Initialise."""
        dict.__init__(self, *args, **kwargs)
        self._l = []

    def push(self, k, v):
        """Store previous state for this key."""
        self._l.append((k, self.get(k, None)))
        if v is None:
            if k in self:
                del self[k]
        else:
            self[k] = v

    def pop(self):
        """Restore previous state for this key."""
        k, v = self._l.pop()
        if v is None:
            if k in self:
                del self[k]
        else:
            self[k] = v
        return k, v



"""Representation of the reference value of a function."""
# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

from ufl.core.operator import Operator
from ufl.core.terminal import FormArgument
from ufl.core.ufl_type import ufl_type


@ufl_type(num_ops=1, is_index_free=True, is_terminal_modifier=True, is_in_reference_frame=True)
class ReferenceValue(Operator):
    """Representation of the reference cell value of a form argument."""

    __slots__ = ()

    def __init__(self, f):
        """Initialise."""
        if not isinstance(f, FormArgument):
            raise ValueError("Can only take reference value of form arguments.")
        Operator.__init__(self, (f,))

    @property
    def ufl_shape(self):
        """Get the UFL shape."""
        return self.ufl_operands[0].ufl_element().reference_value_shape

    def evaluate(self, x, mapping, component, index_values, derivatives=()):
        """Get child from mapping and return the component asked for."""
        raise NotImplementedError()

    def __str__(self):
        """Format as a string."""
        return f"reference_value({self.ufl_operands[0]})"



"""Various expression traversal utilities.

The algorithms here are non-recursive, which is faster than recursion
by a factor of 10 or so because of the function call overhead.
"""

# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Massimiliano Leoni, 2016


def pre_traversal(expr):
    """Yield ``o`` for each tree node ``o`` in *expr*, parent before child."""
    lifo = [expr]
    while lifo:
        expr = lifo.pop()
        yield expr
        for op in expr.ufl_operands:
            lifo.append(op)


def post_traversal(expr):
    """Yield ``o`` for each node ``o`` in *expr*, child before parent."""
    lifo = [(expr, list(reversed(expr.ufl_operands)))]
    while lifo:
        expr, deps = lifo[-1]
        for i, dep in enumerate(deps):
            if dep is not None:
                lifo.append((dep, list(reversed(dep.ufl_operands))))
                deps[i] = None
                break
        else:
            yield expr
            lifo.pop()


def cutoff_post_traversal(expr, cutofftypes):
    """Cut-off post-tranversal.

    Yield ``o`` for each node ``o`` in *expr*, child before parent, but
    skipping subtrees of the cutofftypes.
    """
    lifo = [(expr, list(reversed(expr.ufl_operands)))]
    while lifo:
        expr, deps = lifo[-1]
        if cutofftypes[expr._ufl_typecode_]:
            yield expr
            lifo.pop()
        else:
            for i, dep in enumerate(deps):
                if dep is not None:
                    lifo.append((dep, list(reversed(dep.ufl_operands))))
                    deps[i] = None
                    break
            else:
                yield expr
                lifo.pop()


def unique_pre_traversal(expr, visited=None):
    """Yield ``o`` for each tree node ``o`` in *expr*, parent before child.

    This version only visits each node once.
    """
    if visited is None:
        visited = set()
    lifo = [expr]
    visited.add(expr)

    while lifo:
        expr = lifo.pop()
        yield expr
        for op in expr.ufl_operands:
            if op not in visited:
                lifo.append(op)
                visited.add(op)


def unique_post_traversal(expr, visited=None):
    """Yield ``o`` for each node ``o`` in *expr*, child before parent.

    Never visit a node twice.
    """
    lifo = [(expr, list(expr.ufl_operands))]
    if visited is None:
        visited = set()
    visited.add(expr)
    while lifo:
        expr, deps = lifo[-1]
        for i, dep in enumerate(deps):
            if dep is not None and dep not in visited:
                lifo.append((dep, list(dep.ufl_operands)))
                deps[i] = None
                break
        else:
            yield expr
            visited.add(expr)
            lifo.pop()


def cutoff_unique_post_traversal(expr, cutofftypes, visited=None):
    """Yield ``o`` for each node ``o`` in *expr*, child before parent.

    Never visit a node twice.
    """
    lifo = [(expr, list(reversed(expr.ufl_operands)))]
    if visited is None:
        visited = set()
    while lifo:
        expr, deps = lifo[-1]
        if cutofftypes[expr._ufl_typecode_]:
            yield expr
            visited.add(expr)
            lifo.pop()
        else:
            for i, dep in enumerate(deps):
                if dep is not None and dep not in visited:
                    lifo.append((dep, list(reversed(dep.ufl_operands))))
                    deps[i] = None
                    break
            else:
                yield expr
                visited.add(expr)
                lifo.pop()


def traverse_terminals(expr):
    """Traverse terminals."""
    for op in pre_traversal(expr):
        if op._ufl_is_terminal_:
            yield op


def traverse_unique_terminals(expr, visited=None):
    """Traverse unique terminals."""
    for op in unique_pre_traversal(expr, visited=visited):
        if op._ufl_is_terminal_:
            yield op



"""Test Interpolate object."""

__authors__ = "Nacime Bouziani"
__date__ = "2021-11-19"

import pytest

from ufl import (
    Action,
    Adjoint,
    Argument,
    Coefficient,
    FunctionSpace,
    Mesh,
    TestFunction,
    TrialFunction,
    action,
    adjoint,
    derivative,
    dx,
    grad,
    inner,
    replace,
    triangle,
)
from ufl.algorithms.ad import expand_derivatives
from ufl.algorithms.analysis import (
    extract_arguments,
    extract_arguments_and_coefficients,
    extract_base_form_operators,
    extract_coefficients,
)
from ufl.algorithms.expand_indices import expand_indices
from ufl.core.interpolate import Interpolate
from ufl.finiteelement import FiniteElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1


@pytest.fixture
def domain_2d():
    return Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))


@pytest.fixture
def V1(domain_2d):
    f1 = FiniteElement("CG", triangle, 1, (), identity_pullback, H1)
    return FunctionSpace(domain_2d, f1)


@pytest.fixture
def V2(domain_2d):
    f1 = FiniteElement("CG", triangle, 2, (), identity_pullback, H1)
    return FunctionSpace(domain_2d, f1)


def test_symbolic(V1, V2):
    # Set dual of V2
    V2_dual = V2.dual()

    u = Coefficient(V1)
    vstar = Argument(V2_dual, 0)
    Iu = Interpolate(u, vstar)

    assert Iu == Interpolate(u, V2)
    assert Iu.ufl_function_space() == V2
    assert Iu.argument_slots() == (vstar, u)
    assert Iu.arguments() == (vstar,)
    assert Iu.ufl_operands == (u,)


def test_action_adjoint(V1, V2):
    # Set dual of V2
    V2_dual = V2.dual()
    vstar = Argument(V2_dual, 0)

    u = Coefficient(V1)
    Iu = Interpolate(u, vstar)

    v1 = TrialFunction(V1)
    Iv = Interpolate(v1, vstar)

    assert Iv.argument_slots() == (vstar, v1)
    assert Iv.arguments() == (vstar, v1)

    # -- Action -- #
    v = TestFunction(V1)
    v2 = TrialFunction(V2)
    F = v2 * v * dx
    assert action(Iv, u) == Action(Iv, u)
    assert action(F, Iv) == Action(F, Iv)
    assert action(F, Iu) == Iu * v * dx

    # -- Adjoint -- #
    adjoint(Iv) == Adjoint(Iv)


def test_differentiation(V1, V2):
    u = Coefficient(V1)
    v = TestFunction(V1)

    # Define Interpolate
    Iu = Interpolate(u, V2)

    # -- Differentiate: Interpolate(u, V2) -- #
    uhat = TrialFunction(V1)
    dIu = expand_derivatives(derivative(Iu, u, uhat))

    # dInterpolate(u, v*)/du[uhat] <==> Interpolate(uhat, v*)
    assert dIu == Interpolate(uhat, V2)

    # -- Differentiate: Interpolate(u**2, V2) -- #
    g = u**2
    Ig = Interpolate(g, V2)
    dIg = expand_derivatives(derivative(Ig, u, uhat))
    assert dIg == Interpolate(2 * uhat * u, V2)

    # -- Differentiate: I(u, V2) * v * dx -- #
    F = Iu * v * dx
    Ihat = TrialFunction(Iu.ufl_function_space())
    dFdu = expand_derivatives(derivative(F, u, uhat))
    # Compute dFdu = ∂F/∂u + Action(dFdIu, dIu/du)
    #              = Action(dFdIu, Iu(uhat, v*))
    dFdIu = expand_derivatives(derivative(F, Iu, Ihat))
    assert dFdIu == Ihat * v * dx
    assert dFdu == Action(dFdIu, dIu)

    # -- Differentiate: u * I(u, V2) * v * dx -- #
    F = u * Iu * v * dx
    dFdu = expand_derivatives(derivative(F, u, uhat))
    # Compute dFdu = ∂F/∂u + Action(dFdIu, dIu/du)
    #              = ∂F/∂u + Action(dFdIu, Iu(uhat, v*))
    dFdu_partial = uhat * Iu * v * dx
    dFdIu = Ihat * u * v * dx
    assert dFdu == dFdu_partial + Action(dFdIu, dIu)

    # -- Differentiate (wrt Iu): <Iu, v> + <grad(Iu), grad(v)> - <f, v>
    f = Coefficient(V1)
    F = inner(Iu, v) * dx + inner(grad(Iu), grad(v)) * dx - inner(f, v) * dx
    dFdIu = expand_derivatives(derivative(F, Iu, Ihat))

    # BaseFormOperators are treated as coefficients when a form is differentiated wrt them.
    # -> dFdIu <=> dFdw
    w = Coefficient(V2)
    F = replace(F, {Iu: w})
    dFdw = expand_derivatives(derivative(F, w, Ihat))

    # Need to expand indices to be able to match equal (different MultiIndex used for both).
    assert expand_indices(dFdIu) == expand_indices(dFdw)


def test_extract_base_form_operators(V1, V2):
    u = Coefficient(V1)
    uhat = TrialFunction(V1)
    vstar = Argument(V2.dual(), 0)

    # -- Interpolate(u, V2) -- #
    Iu = Interpolate(u, V2)
    assert extract_arguments(Iu) == [vstar]
    assert extract_arguments_and_coefficients(Iu) == ([vstar], [u])

    F = Iu * dx
    # Form composition: Iu * dx <=> Action(v * dx, Iu(u; v*))
    assert extract_arguments(F) == []
    assert extract_arguments_and_coefficients(F) == ([], [u])

    for e in [Iu, F]:
        assert extract_coefficients(e) == [u]
        assert extract_base_form_operators(e) == [Iu]

    # -- Interpolate(u, V2) -- #
    Iv = Interpolate(uhat, V2)
    assert extract_arguments(Iv) == [vstar, uhat]
    assert extract_arguments_and_coefficients(Iv) == ([vstar, uhat], [])
    assert extract_coefficients(Iv) == []
    assert extract_base_form_operators(Iv) == [Iv]

    # -- Action(v * v2 * dx, Iv) -- #
    v2 = TrialFunction(V2)
    v = TestFunction(V1)
    F = Action(v * v2 * dx, Iv)
    assert extract_arguments(F) == [v, uhat]



"""This module defines the Action class."""
# Copyright (C) 2021 India Marsden
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Nacime Bouziani, 2021-2022.

from itertools import chain

from ufl.algebra import Sum
from ufl.argument import Argument, Coargument
from ufl.coefficient import BaseCoefficient, Coefficient, Cofunction
from ufl.constantvalue import Zero
from ufl.core.base_form_operator import BaseFormOperator
from ufl.core.ufl_type import ufl_type
from ufl.differentiation import CoefficientDerivative
from ufl.form import BaseForm, Form, FormSum, ZeroBaseForm
from ufl.matrix import Matrix

# --- The Action class represents the action of a numerical object that needs
#     to be computed at assembly time ---


@ufl_type()
class Action(BaseForm):
    """UFL base form type: respresents the action of an object on another.

    For example:
        res = Ax
    A would be the first argument, left and x would be the second argument,
    right.

    Action objects will result when the action of an assembled object
    (e.g. a Matrix) is taken. This delays the evaluation of the action until
    assembly occurs.
    """

    __slots__ = (
        "_left",
        "_right",
        "ufl_operands",
        "_repr",
        "_arguments",
        "_coefficients",
        "_domains",
        "_hash",
    )

    def __new__(cls, *args, **kw):
        """Create a new Action."""
        left, right = args

        # Check trivial case
        if left == 0 or right == 0:
            if isinstance(left, Zero):
                # There is no point in checking the action arguments
                # if `left` is a `ufl.Zero` as those objects don't have arguments.
                # We can also not reliably determine the `ZeroBaseForm` arguments.
                return ZeroBaseForm(())
            # Still need to work out the ZeroBaseForm arguments.
            new_arguments, _ = _get_action_form_arguments(left, right)
            return ZeroBaseForm(new_arguments)

        # Coarguments (resp. Argument) from V* to V* (resp. from V to V) are identity matrices,
        # i.e. we have: V* x V -> R (resp. V x V* -> R).
        if isinstance(left, (Coargument, Argument)):
            return right
        if isinstance(right, (Coargument, Argument)):
            return left

        if isinstance(left, (FormSum, Sum)):
            # Action distributes over sums on the LHS
            return FormSum(*[(Action(component, right), 1) for component in left.ufl_operands])
        if isinstance(right, (FormSum, Sum)):
            # Action also distributes over sums on the RHS
            return FormSum(*[(Action(left, component), 1) for component in right.ufl_operands])

        return super(Action, cls).__new__(cls)

    def __init__(self, left, right):
        """Initialise."""
        BaseForm.__init__(self)

        self._left = left
        self._right = right
        self.ufl_operands = (self._left, self._right)
        self._domains = None

        # Check compatibility of function spaces
        _check_function_spaces(left, right)

        self._repr = "Action(%s, %s)" % (repr(self._left), repr(self._right))
        self._hash = None

    def ufl_function_spaces(self):
        """Get the tuple of function spaces of the underlying form."""
        if isinstance(self._right, Form):
            return self._left.ufl_function_spaces()[:-1] + self._right.ufl_function_spaces()[1:]
        elif isinstance(self._right, Coefficient):
            return self._left.ufl_function_spaces()[:-1]

    def left(self):
        """Get left."""
        return self._left

    def right(self):
        """Get right."""
        return self._right

    def _analyze_form_arguments(self):
        """Compute the Arguments of this Action.

        The highest number Argument of the left operand and the lowest number
        Argument of the right operand are consumed by the action.
        """
        self._arguments, self._coefficients = _get_action_form_arguments(self._left, self._right)

    def _analyze_domains(self):
        """Analyze which domains can be found in Action."""
        from ufl.domain import join_domains

        # Collect domains
        self._domains = join_domains(
            chain.from_iterable(e.ufl_domains() for e in self.ufl_operands)
        )

    def ufl_domains(self):
        """Return all domains found in the base form."""
        if self._domains is None:
            self._analyze_domains()
        return self._domains

    def equals(self, other):
        """Check if two Actions are equal."""
        if type(other) is not Action:
            return False
        if self is other:
            return True
        # Make sure we are returning a boolean as left and right equalities can be `ufl.Equation`s
        # if the underlying objects are `ufl.BaseForm`.
        return bool(self._left == other._left) and bool(self._right == other._right)

    def __str__(self):
        """Format as a string."""
        return f"Action({self._left}, {self._right})"

    def __repr__(self):
        """Representation."""
        return self._repr

    def __hash__(self):
        """Hash."""
        if self._hash is None:
            self._hash = hash(("Action", hash(self._right), hash(self._left)))
        return self._hash


def _check_function_spaces(left, right):
    """Check if the function spaces of left and right match."""
    if isinstance(right, CoefficientDerivative):
        # Action differentiation pushes differentiation through
        # right as a consequence of Leibniz formula.
        right, *_ = right.ufl_operands

    # `left` can also be a Coefficient in V (= V**), e.g.
    # `action(Coefficient(V), Cofunction(V.dual()))`.
    left_arg = left.arguments()[-1] if not isinstance(left, Coefficient) else left
    if isinstance(right, (Form, Action, Matrix, ZeroBaseForm)):
        if left_arg.ufl_function_space().dual() != right.arguments()[0].ufl_function_space():
            raise TypeError("Incompatible function spaces in Action")
    elif isinstance(right, (Coefficient, Cofunction, Argument, BaseFormOperator)):
        if left_arg.ufl_function_space() != right.ufl_function_space():
            raise TypeError("Incompatible function spaces in Action")
    # `Zero` doesn't contain any information about the function space.
    # -> Not a problem since Action will get simplified with a
    #    `ZeroBaseForm` which won't take into account the arguments on
    #    the right because of argument contraction.
    # This occurs for:
    # `derivative(Action(A, B), u)` with B is an `Expr` such that dB/du == 0
    # -> `derivative(B, u)` becomes `Zero` when expanding derivatives since B is an Expr.
    elif not isinstance(right, Zero):
        raise TypeError("Incompatible argument in Action: %s" % type(right))


def _get_action_form_arguments(left, right):
    """Perform argument contraction to work out the arguments of Action."""
    coefficients = ()
    # `left` can also be a Coefficient in V (= V**), e.g.
    # `action(Coefficient(V), Cofunction(V.dual()))`.
    left_args = left.arguments()[:-1] if not isinstance(left, Coefficient) else ()
    if isinstance(right, BaseForm):
        arguments = left_args + right.arguments()[1:]
        coefficients += right.coefficients()
    elif isinstance(right, CoefficientDerivative):
        # Action differentiation pushes differentiation through
        # right as a consequence of Leibniz formula.
        from ufl.algorithms.analysis import extract_arguments_and_coefficients

        right_args, right_coeffs = extract_arguments_and_coefficients(right)
        arguments = left_args + tuple(right_args)
        coefficients += tuple(right_coeffs)
    elif isinstance(right, (BaseCoefficient, Zero)):
        arguments = left_args
        # When right is ufl.Zero, Action gets simplified so updating
        # coefficients here doesn't matter
        coefficients += (right,)
    elif isinstance(right, Argument):
        arguments = left_args + (right,)
    else:
        raise TypeError

    if isinstance(left, BaseForm):
        coefficients += left.coefficients()

    return arguments, coefficients



"""FormData class easy for collecting of various data about a form."""

# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Anders Logg, 2008.

from ufl.utils.formatting import estr, lstr, tstr


class FormData(object):
    """Class collecting various information extracted from a Form by calling preprocess."""

    def __init__(self):
        """Create empty form data for given form."""

    def __str__(self):
        """Return formatted summary of form data."""
        types = sorted(self.max_subdomain_ids.keys())
        geometry = (("Geometric dimension", self.geometric_dimension),)
        subdomains = tuple(
            (f"Number of {integral_type} subdomains", self.max_subdomain_ids[integral_type])
            for integral_type in types
        )
        functions = (
            # Arguments
            ("Rank", self.rank),
            ("Arguments", lstr(self.original_form.arguments())),
            # Coefficients
            ("Number of coefficients", self.num_coefficients),
            ("Coefficients", lstr(self.reduced_coefficients)),
            # Elements
            ("Unique elements", estr(self.unique_elements)),
            ("Unique sub elements", estr(self.unique_sub_elements)),
        )
        return tstr(geometry + subdomains + functions)



#
# Author: Martin Sandve Alnes
# Date: 2008-10-03
#
from ufl import Coefficient, FunctionSpace, Mesh, TestFunction, TrialFunction, dx, i, j, triangle
from ufl.finiteelement import FiniteElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1

element = FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1)
domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
space = FunctionSpace(domain, element)

u = TrialFunction(space)
v = TestFunction(space)
w = Coefficient(space)

a = (u[j] * w[i].dx(j) + w[j] * u[i].dx(j)) * v[i] * dx



"""This module defines the ``Expr`` class, the superclass for all expression tree node types in UFL.

NB: A note about other operators not implemented here:

More operators (special functions) on ``Expr`` instances are defined in
``exproperators.py``, as well as the transpose ``A.T`` and spatial derivative
``a.dx(i)``.
This is to avoid circular dependencies between ``Expr`` and its subclasses.
"""
# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Anders Logg, 2008
# Modified by Massimiliano Leoni, 2016

import warnings

from ufl.core.ufl_type import UFLType, update_ufl_type_attributes


class Expr(object, metaclass=UFLType):
    """Base class for all UFL expression types.

    *Instance properties*
        Every ``Expr`` instance will have certain properties.
        The most important ones are ``ufl_operands``, ``ufl_shape``,
        ``ufl_free_indices``, and ``ufl_index_dimensions`` properties.
        Expressions are immutable and hashable.

    *Type traits*
        The ``Expr`` API defines a number of type traits that each subclass
        needs to provide. Most of these are specified indirectly via
        the arguments to the ``ufl_type`` class decorator, allowing UFL
        to do some consistency checks and automate most of the traits
        for most types. Type traits are accessed via a class or
        instance object of the form ``obj._ufl_traitname_``. See the source
        code for description of each type trait.

    *Operators*
        Some Python special functions are implemented in this class,
        some are implemented in subclasses, and some are attached to
        this class in the ``ufl_type`` class decorator.

    *Defining subclasses*
        To define a new expression class, inherit from either
        ``Terminal`` or ``Operator``, and apply the ``ufl_type`` class
        decorator with suitable arguments.  See the docstring of
        ``ufl_type`` for details on its arguments.  Looking at existing
        classes similar to the one you wish to add is a good
        idea. Looking through the comments in the ``Expr`` class and
        ``ufl_type`` to understand all the properties that may need to
        be specified is also a good idea. Note that many algorithms in
        UFL and form compilers will need handlers implemented for each
        new type::.

        .. code-block:: python

            @ufl_type()
            class MyOperator(Operator):
                pass

    *Type collections*
        All ``Expr`` subclasses are collected by ``ufl_type`` in global
        variables available via ``Expr``.

    *Profiling*
        Object creation statistics can be collected by doing

        .. code-block:: python

            Expr.ufl_enable_profiling()
            # ... run some code
            initstats, delstats = Expr.ufl_disable_profiling()

        Giving a list of creation and deletion counts for each typecode.
    """

    # --- Each Expr subclass must define __slots__ or _ufl_noslots_ at
    # --- the top ---
    # This is to freeze member variables for objects of this class and
    # save memory by skipping the per-instance dict.

    __slots__ = ("_hash", "__weakref__")
    # _ufl_noslots_ = True

    # --- Basic object behaviour ---

    def __getnewargs__(self):
        """Get newargs tuple.

        The tuple returned here is passed to as args to cls.__new__(cls, *args).

        This implementation passes the operands, which is () for terminals.

        May be necessary to override if __new__ is implemented in a subclass.
        """
        return self.ufl_operands

    def __init__(self):
        """Initialise."""
        self._hash = None

    # This shows the principal behaviour of the hash function attached
    # in ufl_type:
    # def __hash__(self):
    #     if self._hash is None:
    #         self._hash = self._ufl_compute_hash_()
    #     return self._hash

    # --- Type traits are added to subclasses by the ufl_type class
    # --- decorator ---

    # Note: Some of these are modified after the Expr class definition
    # because Expr is not defined yet at this point.  Note: Boolean
    # type traits that categorize types are mostly set to None for
    # Expr but should be True or False for any non-abstract type.

    # A reference to the UFL class itself.  This makes it possible to
    # do type(f)._ufl_class_ and be sure you get the actual UFL class
    # instead of a subclass from another library.
    _ufl_class_ = None

    # The handler name.  This is the name of the handler function you
    # implement for this type in a multifunction.
    _ufl_handler_name_ = "expr"

    # Number of operands, "varying" for some types, or None if not
    # applicable for abstract types.
    _ufl_num_ops_ = None

    # Type trait: If the type is a literal.
    _ufl_is_literal_ = None

    # Type trait: If the type is classified as a 'terminal modifier',
    # for form compiler use.
    _ufl_is_terminal_modifier_ = None

    # Type trait: If the type is a shaping operator.  Shaping
    # operations include indexing, slicing, transposing, i.e. not
    # introducing computation of a new value.
    _ufl_is_shaping_ = False

    # Type trait: If the type is in reference frame.
    _ufl_is_in_reference_frame_ = None

    # Type trait: If the type is a restriction to a geometric entity.
    _ufl_is_restriction_ = None

    # Type trait: If the type is evaluation in a particular way.
    _ufl_is_evaluation_ = None

    # Type trait: If the type is a differential operator.
    _ufl_is_differential_ = None

    # Type trait: If the type is purely scalar, having no shape or
    # indices.
    _ufl_is_scalar_ = None

    # Type trait: If the type never has free indices.
    _ufl_is_index_free_ = False

    # --- All subclasses must define these object attributes ---

    # Each subclass of Expr is checked to have these properties in
    # ufl_type
    _ufl_required_properties_ = (
        # A tuple of operands, all of them Expr instances.
        "ufl_operands",
        # A tuple of ints, the value shape of the expression.
        "ufl_shape",
        # A tuple of free index counts.
        "ufl_free_indices",
        # A tuple providing the int dimension for each free index.
        "ufl_index_dimensions",
    )

    # Each subclass of Expr is checked to have these methods in
    # ufl_type
    # FIXME: Add more and enable all
    _ufl_required_methods_ = (
        # To compute the hash on demand, this method is called.
        "_ufl_compute_hash_",
        # The data returned from this method is used to compute the
        # signature of a form
        "_ufl_signature_data_",
        # The == operator must be implemented to compare for identical
        # representation, used by set() and dict(). The __hash__
        # operator is added by ufl_type.
        "__eq__",
        # To reconstruct an object of the same type with operands or
        # properties changed.
        "_ufl_expr_reconstruct_",  # Implemented in Operator and Terminal so this should never fail
        "ufl_domains",
        # "ufl_cell",
        # "ufl_domain",
        # "__str__",
        # "__repr__",
    )

    # --- Global variables for collecting all types ---

    # A global dict mapping language_operator_name to the type it
    # produces
    _ufl_language_operators_ = {}

    # List of all terminal modifier types
    _ufl_terminal_modifiers_ = []

    # --- Mechanism for profiling object creation and deletion ---

    # Backup of default init and del
    _ufl_regular__init__ = __init__

    def _ufl_profiling__init__(self):
        """Replacement constructor with object counting."""
        Expr._ufl_regular__init__(self)
        Expr._ufl_obj_init_counts_[self._ufl_typecode_] += 1

    def _ufl_profiling__del__(self):
        """Replacement destructor with object counting."""
        Expr._ufl_obj_del_counts_[self._ufl_typecode_] -= 1

    @staticmethod
    def ufl_enable_profiling():
        """Turn on the object counting mechanism and reset counts to zero."""
        Expr.__init__ = Expr._ufl_profiling__init__
        setattr(Expr, "__del__", Expr._ufl_profiling__del__)
        for i in range(len(Expr._ufl_obj_init_counts_)):
            Expr._ufl_obj_init_counts_[i] = 0
            Expr._ufl_obj_del_counts_[i] = 0

    @staticmethod
    def ufl_disable_profiling():
        """Turn off the object counting mechanism. Return object init and del counts."""
        Expr.__init__ = Expr._ufl_regular__init__
        delattr(Expr, "__del__")
        return (Expr._ufl_obj_init_counts_, Expr._ufl_obj_del_counts_)

    # === Abstract functions that must be implemented by subclasses ===

    # --- Functions for reconstructing expression ---

    def _ufl_expr_reconstruct_(self, *operands):
        """Return a new object of the same type with new operands."""
        raise NotImplementedError(self.__class__._ufl_expr_reconstruct_)

    # --- Functions for geometric properties of expression ---

    def ufl_domains(self):
        """Return all domains this expression is defined on."""
        warnings.warn(
            "Expr.ufl_domains() is deprecated, please use extract_domains(expr) instead.",
            DeprecationWarning,
        )
        from ufl.domain import extract_domains

        return extract_domains(self)

    def ufl_domain(self):
        """Return the single unique domain this expression is defined on, or throw an error."""
        warnings.warn(
            "Expr.ufl_domain() is deprecated, please use extract_unique_domain(expr) instead.",
            DeprecationWarning,
        )
        from ufl.domain import extract_unique_domain

        return extract_unique_domain(self)

    # --- Functions for float evaluation ---

    def evaluate(self, x, mapping, component, index_values):
        """Evaluate expression at given coordinate with given values for terminals."""
        raise ValueError(f"Symbolic evaluation of {self._ufl_class_.__name__} not available.")

    def _ufl_evaluate_scalar_(self):
        if self.ufl_shape or self.ufl_free_indices:
            raise TypeError("Cannot evaluate a nonscalar expression to a scalar value.")
        return self(())  # No known x

    def __float__(self):
        """Try to evaluate as scalar and cast to float."""
        try:
            v = float(self._ufl_evaluate_scalar_())
        except Exception:
            v = NotImplemented
        return v

    def __complex__(self):
        """Try to evaluate as scalar and cast to complex."""
        try:
            v = complex(self._ufl_evaluate_scalar_())
        except TypeError:
            v = NotImplemented
        return v

    def __bool__(self):
        """By default, all Expr are nonzero/False."""
        return True

    def __nonzero__(self):
        """By default, all Expr are nonzero/False."""
        return self.__bool__()

    @staticmethod
    def _ufl_coerce_(value):
        """Convert any value to a UFL type."""
        # Quick skip for most types
        if isinstance(value, Expr):
            return value

        # Conversion from non-ufl types
        # (the _ufl_from_*_ functions are attached to Expr by ufl_type)
        ufl_from_type = "_ufl_from_{0}_".format(value.__class__.__name__)
        return getattr(Expr, ufl_from_type)(value)

        # if hasattr(Expr, ufl_from_type):
        #     return getattr(Expr, ufl_from_type)(value)
        # Fail gracefully if no valid type conversion found
        # raise TypeError("Cannot convert a {0.__class__.__name__} to UFL type.".format(value))

    # --- Special functions for string representations ---

    # All subclasses must implement _ufl_signature_data_
    def _ufl_signature_data_(self, renumbering):
        """Return data that uniquely identifies form compiler relevant aspects of this object."""
        raise NotImplementedError(self.__class__._ufl_signature_data_)

    # All subclasses must implement __repr__
    def __repr__(self):
        """Return string representation this object can be reconstructed from."""
        raise NotImplementedError(self.__class__.__repr__)

    # All subclasses must implement __str__
    def __str__(self):
        """Return pretty print string representation of this object."""
        raise NotImplementedError(self.__class__.__str__)

    def _ufl_err_str_(self):
        """Return a short string to represent this Expr in an error message."""
        return f"<{self._ufl_class_.__name__} id={id(self)}>"

    # --- Special functions used for processing expressions ---

    def __eq__(self, other):
        """Checks whether the two expressions are represented the exact same way.

        This does not check if the expressions are
        mathematically equal or equivalent! Used by sets and dicts.
        """
        raise NotImplementedError(self.__class__.__eq__)

    def __len__(self):
        """Length of expression. Used for iteration over vector expressions."""
        s = self.ufl_shape
        if len(s) == 1:
            return s[0]
        raise NotImplementedError("Cannot take length of non-vector expression.")

    def __iter__(self):
        """Iteration over vector expressions."""
        for i in range(len(self)):
            yield self[i]

    def __floordiv__(self, other):
        """UFL does not support integer division."""
        raise NotImplementedError(self.__class__.__floordiv__)

    def __pos__(self):
        """Unary + is a no-op."""
        return self

    def __round__(self, n=None):
        """Round to nearest integer or to nearest nth decimal."""
        try:
            val = float(self._ufl_evaluate_scalar_())
            val = round(val, n)
        except TypeError:
            val = complex(self._ufl_evaluate_scalar_())
            val = round(val.real, n) + round(val.imag, n) * 1j
        except TypeError:
            val = NotImplemented
        return val


# Initializing traits here because Expr is not defined in the class
# declaration
Expr._ufl_class_ = Expr

# Update Expr with metaclass properties (e.g. typecode or handler name)
# Explicitly done here instead of using `@ufl_type` to avoid circular imports.
update_ufl_type_attributes(Expr)


def ufl_err_str(expr):
    """Return a UFL error string."""
    if hasattr(expr, "_ufl_err_str_"):
        return expr._ufl_err_str_()
    else:
        return repr(expr)



# Copyright (C) 2006-2009 Anders Logg and Marie Rognes
#
# This file is part of UFL.
#
# UFL is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# UFL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with UFL. If not, see <http://www.gnu.org/licenses/>.
#
# Modified by Martin Sandve Alnes, 2009
#
# Last changed: 2009-01-12
#
# The bilinear form a(v, u) and linear form L(v) for
# a mixed formulation of Poisson's equation with BDM
# (Brezzi-Douglas-Marini) elements.
#
from ufl import (
    Coefficient,
    FunctionSpace,
    Mesh,
    TestFunctions,
    TrialFunctions,
    div,
    dot,
    dx,
    triangle,
)
from ufl.finiteelement import FiniteElement, MixedElement
from ufl.pullback import contravariant_piola, identity_pullback
from ufl.sobolevspace import H1, HDiv

cell = triangle
BDM1 = FiniteElement("Brezzi-Douglas-Marini", cell, 1, (2,), contravariant_piola, HDiv)
DG0 = FiniteElement("Discontinuous Lagrange", cell, 0, (), identity_pullback, H1)

element = MixedElement([BDM1, DG0])
domain = Mesh(FiniteElement("Lagrange", cell, 1, (2,), identity_pullback, H1))
space = FunctionSpace(domain, element)
dg0_space = FunctionSpace(domain, DG0)

(tau, w) = TestFunctions(space)
(sigma, u) = TrialFunctions(space)

f = Coefficient(dg0_space)

a = (dot(tau, sigma) - div(tau) * u + w * div(sigma)) * dx
L = w * f * dx



"""Functions to check the validity of forms."""

# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Anders Logg, 2008-2009.
# Modified by Mehdi Nikbakht, 2010.

from ufl.algorithms.check_restrictions import check_restrictions

# UFL algorithms
from ufl.algorithms.traversal import iter_expressions
from ufl.argument import Argument
from ufl.coefficient import Coefficient
from ufl.constantvalue import is_true_ufl_scalar

# UFL classes
from ufl.core.expr import ufl_err_str
from ufl.corealg.traversal import traverse_unique_terminals
from ufl.domain import extract_unique_domain
from ufl.form import Form


def validate_form(
    form,
):  # TODO: Can we make this return a list of errors instead of raising exception?
    """Performs all implemented validations on a form. Raises exception if something fails."""
    errors = []

    if not isinstance(form, Form):
        raise ValueError(f"Validation failed, not a Form:\n{ufl_err_str(form)}")

    # FIXME: There's a bunch of other checks we should do here.

    # FIXME: Add back check for multilinearity
    # Check that form is multilinear
    # if not is_multilinear(form):
    #     errors.append("Form is not multilinear in arguments.")

    # FIXME DOMAIN: Add check for consistency between domains somehow
    domains = set(
        extract_unique_domain(t)
        for e in iter_expressions(form)
        for t in traverse_unique_terminals(e)
    ) - {None}
    if not domains:
        errors.append("Missing domain definition in form.")

    # Check that cell is the same everywhere
    cells = set(dom.ufl_cell() for dom in domains) - {None}
    if not cells:
        errors.append("Missing cell definition in form.")
    elif len(cells) > 1:
        errors.append(f"Multiple cell definitions in form: {cells}")

    # Check that no Coefficient or Argument instance have the same
    # count unless they are the same
    coefficients = {}
    arguments = {}
    for e in iter_expressions(form):
        for f in traverse_unique_terminals(e):
            if isinstance(f, Coefficient):
                c = f.count()
                if c in coefficients:
                    g = coefficients[c]
                    if f is not g:
                        errors.append(f"Found different Coefficients with same count: {f} and {g}.")
                else:
                    coefficients[c] = f

            elif isinstance(f, Argument):
                n = f.number()
                p = f.part()
                if (n, p) in arguments:
                    g = arguments[(n, p)]
                    if f is not g:
                        if n == 0:
                            msg = "TestFunctions"
                        elif n == 1:
                            msg = "TrialFunctions"
                        else:
                            msg = "Arguments with same number and part"
                        msg = f"Found different {msg}: {f!r} and {g!r}."
                        errors.append(msg)
                else:
                    arguments[(n, p)] = f

    # Check that all integrands are scalar
    for expression in iter_expressions(form):
        if not is_true_ufl_scalar(expression):
            errors.append("Found non-scalar integrand expression: {ufl_err_str(expression)}\n")

    # Check that restrictions are permissible
    for integral in form.integrals():
        # Only allow restrictions on interior facet integrals and
        # surface measures
        if integral.integral_type().startswith("interior_facet"):
            check_restrictions(integral.integrand(), True)
        else:
            check_restrictions(integral.integrand(), False)

    # Raise exception with all error messages
    # TODO: Return errors list instead, need to collect messages from
    # all validations above first.
    if errors:
        raise ValueError("Found errors in validation of form:\n" + "\n\n".join(errors))



"""Utilities for transforming complete Forms into new related Forms."""

# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Anders Logg, 2008-2009.
# Modified by Garth N. Wells, 2010.
# Modified by Marie E. Rognes, 2010.

import warnings
from logging import debug

from ufl.algebra import Conj

# Other algorithms:
from ufl.algorithms.map_integrands import map_integrands
from ufl.algorithms.replace import replace
from ufl.algorithms.transformer import Transformer
from ufl.argument import Argument
from ufl.coefficient import Coefficient
from ufl.constantvalue import Zero

# All classes:
from ufl.core.expr import ufl_err_str


# FIXME: Don't use this below, it makes partextracter more expensive than necessary
def _expr_has_terminal_types(expr, ufl_types):
    """Check if an expression has terminal types."""
    input = [expr]
    while input:
        e = input.pop()
        ops = e.ufl_operands
        if ops:
            input.extend(ops)
        elif isinstance(e, ufl_types):
            return True
    return False


def zero_expr(e):
    """Create a zero expression."""
    return Zero(e.ufl_shape, e.ufl_free_indices, e.ufl_index_dimensions)


class PartExtracter(Transformer):
    """PartExtracter extracts those parts of a form that contain the given argument(s)."""

    def __init__(self, arguments):
        """Initialise."""
        Transformer.__init__(self)
        self._want = set(arguments)

    def expr(self, x):
        """Apply to expr.

        The default is a nonlinear operator not accepting any Arguments among its children.
        """
        if _expr_has_terminal_types(x, Argument):
            raise ValueError(f"Found Argument in {ufl_err_str(x)}, this is an invalid expression.")
        return (x, set())

    # Terminals that are not Variables or Arguments behave as default
    # expr-s.
    terminal = expr

    def variable(self, x):
        """Return relevant parts of this variable."""
        # Extract parts/provides from this variable's expression
        expression, label = x.ufl_operands
        part, provides = self.visit(expression)

        # If the extracted part is zero or we provide more than we
        # want, return zero
        if isinstance(part, Zero) or (provides - self._want):
            return (zero_expr(x), set())

        # Reuse varible if possible (or reconstruct from part)
        x = self.reuse_if_possible(x, part, label)

        return (x, provides)

    def argument(self, x):
        """Return itself unless itself provides too much."""
        # An argument provides itself
        provides = {x}

        # If we provide more than we want, return zero
        if provides - self._want:
            return (zero_expr(x), set())

        return (x, provides)

    def sum(self, x):
        """Return the terms that might eventually yield the correct parts(!).

        The logic required for sums is a bit elaborate:

        A sum may contain terms providing different arguments. We
        should return (a sum of) a suitable subset of these
        terms. Those should all provide the same arguments.

        For each term in a sum, there are 2 simple possibilities:

        1a) The relevant part of the term is zero -> skip.
        1b) The term provides more arguments than we want -> skip

        2) If all terms fall into the above category, we can just
        return zero.

        Any remaining terms may provide exactly the arguments we want,
        or fewer. This is where things start getting interesting.

        3) Bottom-line: if there are terms with providing different
        arguments -- provide terms that contain the most arguments. If
        there are terms providing different sets of same size -> throw
        error (e.g. Argument(-1) + Argument(-2)).
        """
        parts_that_provide = {}

        # 1. Skip terms that provide too much
        original_terms = x.ufl_operands
        assert len(original_terms) == 2
        for term in original_terms:
            # Visit this term in the sum
            part, term_provides = self.visit(term)

            # If this part is zero or it provides more than we want,
            # skip it
            if isinstance(part, Zero) or (term_provides - self._want):
                continue

            # Add the contributions from this part to temporary list
            term_provides = frozenset(term_provides)
            if term_provides in parts_that_provide:
                parts_that_provide[term_provides] += [part]
            else:
                parts_that_provide[term_provides] = [part]

        # 2. If there are no remaining terms, return zero
        if not parts_that_provide:
            return (zero_expr(x), set())

        # 3. Return the terms that provide the biggest set
        most_provided = frozenset()
        for provideds, parts in parts_that_provide.items():  # TODO: Just sort instead?
            # Throw error if size of sets are equal (and not zero)
            if len(provideds) == len(most_provided) and len(most_provided):
                raise ValueError("Don't know what to do with sums with different Arguments.")

            if provideds > most_provided:
                most_provided = provideds

        terms = parts_that_provide[most_provided]
        if len(terms) == 2:
            x = self.reuse_if_possible(x, *terms)
        else:
            (x,) = terms

        return (x, most_provided)

    def product(self, x, *ops):
        """Apply to product.

        Note: Product is a visit-children-first handler. ops are
        the visited factors.
        """
        provides = set()
        factors = []

        for factor, factor_provides in ops:
            # If any factor is zero, return
            if isinstance(factor, Zero):
                return (zero_expr(x), set())

            # Add factor to factors and extend provides
            factors.append(factor)
            provides = provides | factor_provides

            # If we provide more than we want, return zero
            if provides - self._want:
                return (zero_expr(x), provides)

        # Reuse product if possible (or reconstruct from factors)
        x = self.reuse_if_possible(x, *factors)

        return (x, provides)

    # inner, outer and dot all behave as product
    inner = product
    outer = product
    dot = product

    def division(self, x):
        """Return parts_of_numerator/denominator."""
        # Get numerator and denominator
        numerator, denominator = x.ufl_operands

        # Check for Arguments in the denominator
        if _expr_has_terminal_types(denominator, Argument):
            raise ValueError(
                f"Found Argument in denominator of {ufl_err_str(x)}, this is an invalid expression."
            )

        # Visit numerator
        numerator_parts, provides = self.visit(numerator)

        # If numerator is zero, return zero. (No need to check whether
        # it provides too much, already checked by visit.)
        if isinstance(numerator_parts, Zero):
            return (zero_expr(x), set())

        # Reuse x if possible, otherwise reconstruct from (parts of)
        # numerator and denominator
        x = self.reuse_if_possible(x, numerator_parts, denominator)

        return (x, provides)

    def linear_operator(self, x, arg):
        """Apply to linear_operator.

        A linear operator with a single operand accepting arity > 0,
        providing whatever Argument its operand does.
        """
        # linear_operator is a visit-children-first handler. Handled
        # arguments are in arg.
        part, provides = arg

        # Discard if part is zero. (No need to check whether we
        # provide too much, already checked by children.)
        if isinstance(part, Zero):
            return (zero_expr(x), set())

        x = self.reuse_if_possible(x, part)

        return (x, provides)

    # Positive and negative restrictions behave as linear operators
    positive_restricted = linear_operator
    negative_restricted = linear_operator

    # Cell and facet average are linear operators
    cell_avg = linear_operator
    facet_avg = linear_operator

    # Grad is a linear operator
    grad = linear_operator

    # Conj, Real, Imag are linear operators
    conj = linear_operator
    real = linear_operator
    imag = linear_operator

    def linear_indexed_type(self, x):
        """Return parts of expression belonging to this indexed expression."""
        expression, index = x.ufl_operands
        part, provides = self.visit(expression)

        # Return zero if extracted part is zero. (The expression
        # should already have checked if it provides too much.)
        if isinstance(part, Zero):
            return (zero_expr(x), set())

        # Reuse x if possible (or reconstruct by indexing part)
        x = self.reuse_if_possible(x, part, index)

        return (x, provides)

    # All of these indexed thingies behave as a linear_indexed_type
    indexed = linear_indexed_type
    index_sum = linear_indexed_type
    component_tensor = linear_indexed_type

    def list_tensor(self, x, *ops):
        """Apply to list_tensor."""
        # list_tensor is a visit-children-first handler. ops contains
        # the visited operands with their provides. (It follows that
        # none of the visited operands provide more than wanted.)

        # Extract the most arguments provided by any of the components
        most_provides = ops[0][1]
        for component, provides in ops:
            if provides - most_provides:
                most_provides = provides

        # Check that all components either provide the same arguments
        # or vanish. (This check is here b/c it is not obvious what to
        # return if the components provide different arguments, at
        # least with the current transformer design.)
        for component, provides in ops:
            if provides != most_provides and not isinstance(component, Zero):
                raise ValueError(
                    "PartExtracter does not know how to handle list_tensors with "
                    "non-zero components providing fewer arguments"
                )

        # Return components
        components = [op[0] for op in ops]
        x = self.reuse_if_possible(x, *components)

        return (x, most_provides)


def compute_form_with_arity(form, arity, arguments=None):
    """Compute parts of form of given arity."""
    # Extract all arguments in form
    if arguments is None:
        arguments = form.arguments()

    parts = [arg.part() for arg in arguments]
    if set(parts) - {None}:
        raise ValueError("compute_form_with_arity cannot handle parts.")

    if len(arguments) < arity:
        warnings.warn(f"Form has no parts with arity {arity}.")
        return 0 * form

    # Assuming that the form is not a sum of terms
    # that depend on different arguments, e.g. (u+v)*dx
    # would result in just v*dx. But that doesn't make
    # any sense anyway.
    sub_arguments = set(arguments[:arity])
    pe = PartExtracter(sub_arguments)

    def _transform(e):
        e, provides = pe.visit(e)
        if provides == sub_arguments:
            return e
        return Zero()

    return map_integrands(_transform, form)


def compute_form_arities(form):
    """Return set of arities of terms present in form."""
    # Extract all arguments present in form
    arguments = form.arguments()

    parts = [arg.part() for arg in arguments]
    if set(parts) - {None}:
        raise ValueError("compute_form_arities cannot handle parts.")

    arities = set()
    for arity in range(len(arguments) + 1):
        # Compute parts with arity "arity"
        parts = compute_form_with_arity(form, arity, arguments)

        # Register arity if "parts" does not vanish
        if parts and parts.integrals():
            arities.add(arity)

    return arities


def compute_form_lhs(form):
    """Compute the left hand side of a form.

    Example:
        a = u*v*dx + f*v*dx
        a = lhs(a) -> u*v*dx
    """
    return compute_form_with_arity(form, 2)


def compute_form_rhs(form):
    """Compute the right hand side of a form.

    Example:
        a = u*v*dx + f*v*dx
        L = rhs(a) -> -f*v*dx
    """
    return -compute_form_with_arity(form, 1)


def compute_form_functional(form):
    """Compute the functional part of a form, that is the terms independent of Arguments.

    (Used for testing, not sure if it's useful for anything?)
    """
    return compute_form_with_arity(form, 0)


def compute_form_action(form, coefficient):
    """Compute the action of a form on a Coefficient.

    This works simply by replacing the last Argument
    with a Coefficient on the same function space (element).
    The form returned will thus have one Argument less
    and one additional Coefficient at the end if no
    Coefficient has been provided.
    """
    # TODO: Check whatever makes sense for coefficient

    # Extract all arguments
    arguments = form.arguments()

    parts = [arg.part() for arg in arguments]
    if set(parts) - {None}:
        raise ValueError("compute_form_action cannot handle parts.")

    # Pick last argument (will be replaced)
    u = arguments[-1]

    fs = u.ufl_function_space()
    if coefficient is None:
        coefficient = Coefficient(fs)
    elif coefficient.ufl_function_space() != fs:
        debug("Computing action of form on a coefficient in a different function space.")
    return replace(form, {u: coefficient})


def compute_energy_norm(form, coefficient):
    """Compute the a-norm of a Coefficient given a form a.

    This works simply by replacing the two Arguments
    with a Coefficient on the same function space (element).
    The Form returned will thus be a functional with no
    Arguments, and one additional Coefficient at the
    end if no coefficient has been provided.
    """
    from ufl.formoperators import action  # Delayed import to avoid circularity

    arguments = form.arguments()

    parts = [arg.part() for arg in arguments]
    if set(parts) - {None}:
        raise ValueError("compute_energy_norm cannot handle parts.")

    if len(arguments) != 2:
        raise ValueError("Expecting bilinear form.")
    v, u = arguments
    U = u.ufl_function_space()
    V = v.ufl_function_space()
    if U != V:
        raise ValueError(
            f"Expecting equal finite elements for test and trial functions, got '{U}' and '{V}'."
        )
    if coefficient is None:
        coefficient = Coefficient(V)
    else:
        if coefficient.ufl_function_space() != U:
            raise ValueError(
                "Trying to compute action of form on a "
                "coefficient in an incompatible element space."
            )
    return action(action(form, coefficient), coefficient)


def compute_form_adjoint(form, reordered_arguments=None):
    """Compute the adjoint of a bilinear form.

    This works simply by swapping the number and part of the two arguments,
    but keeping their elements and places in the integrand expressions.
    """
    arguments = form.arguments()

    parts = [arg.part() for arg in arguments]
    if set(parts) - {None}:
        raise ValueError("compute_form_adjoint cannot handle parts.")

    if len(arguments) != 2:
        raise ValueError("Expecting bilinear form.")

    v, u = arguments
    if v.number() >= u.number():
        raise ValueError("Mistaken assumption in code!")

    if reordered_arguments is None:
        reordered_u = Argument(u.ufl_function_space(), number=v.number(), part=v.part())
        reordered_v = Argument(v.ufl_function_space(), number=u.number(), part=u.part())
    else:
        reordered_u, reordered_v = reordered_arguments

    if reordered_u.number() >= reordered_v.number():
        raise ValueError("Ordering of new arguments is the same as the old arguments!")

    if reordered_u.part() != v.part():
        raise ValueError("Ordering of new arguments is the same as the old arguments!")
    if reordered_v.part() != u.part():
        raise ValueError("Ordering of new arguments is the same as the old arguments!")

    if reordered_u.ufl_function_space() != u.ufl_function_space():
        raise ValueError("Element mismatch between new and old arguments (trial functions).")
    if reordered_v.ufl_function_space() != v.ufl_function_space():
        raise ValueError("Element mismatch between new and old arguments (test functions).")

    return map_integrands(Conj, replace(form, {v: reordered_v, u: reordered_u}))



# Copyright (C) 2005-2009 Anders Logg
#
# This file is part of UFL.
#
# UFL is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# UFL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with UFL. If not, see <http://www.gnu.org/licenses/>.
#
# Modified by Martin Sandve Alnes, 2009
#
# The bilinear form a(v, u1) and linear form L(v) for
# one backward Euler step with the heat equation.
#
from ufl import (
    Coefficient,
    Constant,
    FunctionSpace,
    Mesh,
    TestFunction,
    TrialFunction,
    dot,
    dx,
    grad,
    triangle,
)
from ufl.finiteelement import FiniteElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1

cell = triangle
element = FiniteElement("Lagrange", cell, 1, (), identity_pullback, H1)
domain = Mesh(FiniteElement("Lagrange", cell, 1, (2,), identity_pullback, H1))
space = FunctionSpace(domain, element)

v = TestFunction(space)  # Test function
u1 = TrialFunction(space)  # Value at t_n
u0 = Coefficient(space)  # Value at t_n-1
c = Coefficient(space)  # Heat conductivity
f = Coefficient(space)  # Heat source
k = Constant(domain)  # Time step

a = v * u1 * dx + k * c * dot(grad(v), grad(u1)) * dx
L = v * u0 * dx + k * v * f * dx



"""Remove conj, real, and imag nodes from a form."""

from ufl.algorithms.map_integrands import map_integrand_dags
from ufl.constantvalue import ComplexValue
from ufl.corealg.multifunction import MultiFunction


class ComplexNodeRemoval(MultiFunction):
    """Replaces complex operator nodes with their children."""

    expr = MultiFunction.reuse_if_untouched

    def conj(self, o, a):
        """Apply to conj."""
        return a

    def real(self, o, a):
        """Apply to real."""
        return a

    def imag(self, o, a):
        """Apply to imag."""
        raise ValueError("Unexpected imag in real expression.")

    def terminal(self, t, *ops):
        """Apply to terminal."""
        if isinstance(t, ComplexValue):
            raise ValueError("Unexpected complex value in real expression.")
        else:
            return t


def remove_complex_nodes(expr):
    """Replaces complex operator nodes with their children.

    This is called during compute_form_data if the compiler wishes to
    compile real-valued forms. In essence this strips all trace of
    complex support from the preprocessed form.
    """
    return map_integrand_dags(ComplexNodeRemoval(), expr)



__authors__ = "Martin Sandve Alnæs"
__date__ = "2008-09-28 -- 2008-09-28"

import os
from glob import glob

import pytest

from ufl.algorithms import load_ufl_file, validate_form

demodir = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "demo"))


def get_demo_filenames():
    filenames = sorted(
        set(glob(os.path.join(demodir, "*.py"))) - set(glob(os.path.join(demodir, "_*.py")))
    )
    return filenames


@pytest.mark.parametrize("filename", get_demo_filenames())
def test_demo_files(filename):
    "Check each form in each file with validate_form."
    data = load_ufl_file(filename)
    for form in data.forms:
        validate_form(form)



"""Utility objects for pretty syntax in user code."""
# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Anders Logg, 2008
# Modified by Kristian Oelgaard, 2009

from ufl.cell import Cell
from ufl.core.multiindex import indices
from ufl.measure import Measure, integral_type_to_measure_name

# Default indices
i, j, k, l = indices(4)  # noqa: E741
p, q, r, s = indices(4)

for integral_type, measure_name in integral_type_to_measure_name.items():
    globals()[measure_name] = Measure(integral_type)

# TODO: Firedrake hack, remove later
ds_tb = ds_b + ds_t  # noqa: F821

# Default measure dX including both uncut and cut cells
dX = dx + dC  # noqa: F821

# Create objects for builtin known cell types
vertex = Cell("vertex")
interval = Cell("interval")
triangle = Cell("triangle")
tetrahedron = Cell("tetrahedron")
prism = Cell("prism")
pyramid = Cell("pyramid")
quadrilateral = Cell("quadrilateral")
hexahedron = Cell("hexahedron")
tesseract = Cell("tesseract")
pentatope = Cell("pentatope")

# Facet is just a dummy declaration for RestrictedElement
facet = "facet"



#
# Author: Martin Sandve Alnes
# Date: 2008-10-28
#
from ufl import Coefficient, FunctionSpace, Mesh, derivative, dx, triangle
from ufl.finiteelement import FiniteElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1

element = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
space = FunctionSpace(domain, element)

u = Coefficient(space)

# L2 norm
M = u**2 / 2 * dx
# source vector
L = derivative(M, u)
# mass matrix
a = derivative(L, u)



"""Tools to strip away and reattach coordinate derivatives.

This is used in compute_form_data.
"""

# Copyright (C) 2018 Florian Wechsung
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

from ufl.classes import Integral
from ufl.corealg.map_dag import map_expr_dags
from ufl.corealg.multifunction import MultiFunction
from ufl.differentiation import CoordinateDerivative


class CoordinateDerivativeIsOutermostChecker(MultiFunction):
    """Traverses the tree to make sure that CoordinateDerivatives are only on the outside.

    The visitor returns False as long as no CoordinateDerivative has been seen.
    """

    def multi_index(self, o):
        """Apply to multi_index."""
        return False

    def terminal(self, o):
        """Apply to terminal."""
        return False

    def expr(self, o, *operands):
        """Apply to expr.

        If we have already seen a CoordinateDerivative, then no other
        expressions apart from more CoordinateDerivatives are allowed to wrap
        around it.
        """
        if any(operands):
            raise ValueError("CoordinateDerivative(s) must be outermost")
        return False

    def coordinate_derivative(self, o, expr, *_):
        """Apply to coordinate derivative."""
        return True


def strip_coordinate_derivatives(integrals):
    """Strip coordinate derivatives."""
    if isinstance(integrals, list):
        if len(integrals) == 0:
            return integrals, None
        stripped_integrals_and_cds = []
        for integral in integrals:
            (si, cd) = strip_coordinate_derivatives(integral)
            stripped_integrals_and_cds.append((si, cd))
        return stripped_integrals_and_cds

    elif isinstance(integrals, Integral):
        integral = integrals
        integrand = integral.integrand()
        checker = CoordinateDerivativeIsOutermostChecker()
        map_expr_dags(checker, [integrand])
        coordinate_derivatives = []

        def take_top_coordinate_derivatives(o):
            """Get all coordinate derivatives and store them.

            So we can apply them later again.
            """
            o_ = o.ufl_operands
            if isinstance(o, CoordinateDerivative):
                coordinate_derivatives.append((o_[1], o_[2], o_[3]))
                return take_top_coordinate_derivatives(o_[0])
            else:
                return o

        newintegrand = take_top_coordinate_derivatives(integrand)
        return (integral.reconstruct(integrand=newintegrand), coordinate_derivatives)

    else:
        raise ValueError(f"Invalid type {integrals.__class__.__name__}")


def attach_coordinate_derivatives(integral, coordinate_derivatives):
    """Attach coordinate derivatives."""
    if coordinate_derivatives is None:
        return integral

    if isinstance(integral, Integral):
        integrand = integral.integrand()
        # apply the stored coordinate derivatives back onto the integrand
        for tup in reversed(coordinate_derivatives):
            integrand = CoordinateDerivative(integrand, tup[0], tup[1], tup[2])
        return integral.reconstruct(integrand=integrand)
    else:
        raise ValueError(f"Invalid type {integral.__class__.__name__}")



__authors__ = "India Marsden"
__date__ = "2020-12-28"

import pytest

from ufl import (
    Action,
    Adjoint,
    Argument,
    Coargument,
    Coefficient,
    Cofunction,
    FormSum,
    FunctionSpace,
    Matrix,
    Mesh,
    MixedFunctionSpace,
    TestFunction,
    TrialFunction,
    action,
    adjoint,
    derivative,
    dx,
    inner,
    interval,
    tetrahedron,
    triangle,
)
from ufl.algorithms.ad import expand_derivatives
from ufl.constantvalue import Zero
from ufl.duals import is_dual, is_primal
from ufl.finiteelement import FiniteElement
from ufl.form import ZeroBaseForm
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1


def test_mixed_functionspace(self):
    # Domains
    domain_3d = Mesh(FiniteElement("Lagrange", tetrahedron, 1, (3,), identity_pullback, H1))
    domain_2d = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    domain_1d = Mesh(FiniteElement("Lagrange", interval, 1, (1,), identity_pullback, H1))
    # Finite elements
    f_1d = FiniteElement("Lagrange", interval, 1, (), identity_pullback, H1)
    f_2d = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
    f_3d = FiniteElement("Lagrange", tetrahedron, 1, (), identity_pullback, H1)
    # Function spaces
    V_3d = FunctionSpace(domain_3d, f_3d)
    V_2d = FunctionSpace(domain_2d, f_2d)
    V_1d = FunctionSpace(domain_1d, f_1d)

    # MixedFunctionSpace = V_3d x V_2d x V_1d
    V = MixedFunctionSpace(V_3d, V_2d, V_1d)
    # Check sub spaces
    assert is_primal(V_3d)
    assert is_primal(V_2d)
    assert is_primal(V_1d)
    assert is_primal(V)

    # Get dual of V_3
    V_dual = V_3d.dual()

    #  Test dual functions on MixedFunctionSpace = V_dual x V_2d x V_1d
    V = MixedFunctionSpace(V_dual, V_2d, V_1d)
    V_mixed_dual = MixedFunctionSpace(V_dual, V_2d.dual(), V_1d.dual())

    assert is_dual(V_dual)
    assert not is_dual(V)
    assert is_dual(V_mixed_dual)


def test_dual_coefficients():
    domain_2d = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    f_2d = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
    V = FunctionSpace(domain_2d, f_2d)
    V_dual = V.dual()

    v = Coefficient(V, count=1)
    u = Coefficient(V_dual, count=1)
    w = Cofunction(V_dual)

    assert is_primal(v)
    assert not is_dual(v)

    assert is_dual(u)
    assert not is_primal(u)

    assert is_dual(w)
    assert not is_primal(w)

    with pytest.raises(ValueError):
        Cofunction(V)


def test_dual_arguments():
    domain_2d = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    f_2d = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
    V = FunctionSpace(domain_2d, f_2d)
    V_dual = V.dual()

    v = Argument(V, 1)
    u = Argument(V_dual, 2)
    w = Coargument(V_dual, 3)

    assert is_primal(v)
    assert not is_dual(v)

    assert is_dual(u)
    assert not is_primal(u)

    assert is_dual(w)
    assert not is_primal(w)

    with pytest.raises(ValueError):
        Coargument(V, 4)


def test_addition():
    domain_2d = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    f_2d = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
    V = FunctionSpace(domain_2d, f_2d)
    V_dual = V.dual()

    fvector_2d = FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1)
    W = FunctionSpace(domain_2d, fvector_2d)

    u = TrialFunction(V)
    v = TestFunction(V)

    # linear 1-form
    L = v * dx
    a = Cofunction(V_dual)
    res = L + a
    assert isinstance(res, FormSum)
    assert res

    L = u * v * dx
    a = Matrix(V, V)
    res = L + a
    assert isinstance(res, FormSum)
    assert res

    # Check BaseForm._add__ simplification
    res += ZeroBaseForm((v, u))
    assert res == a + L
    # Check Form._add__ simplification
    L += ZeroBaseForm((v,))
    assert L == u * v * dx
    # Check BaseForm._add__ simplification
    res = ZeroBaseForm((v, u))
    res += a
    assert res == a
    # Check __neg__
    res = L
    res -= ZeroBaseForm((v,))
    assert res == L

    # Simplification with respect to ufl.Zero
    a_W = Matrix(W, W)
    res = a_W + Zero(W.value_shape)
    assert res == a_W


def test_scalar_mult():
    domain_2d = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    f_2d = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
    V = FunctionSpace(domain_2d, f_2d)
    V_dual = V.dual()

    # linear 1-form
    a = Cofunction(V_dual)
    res = 2 * a
    assert isinstance(res, FormSum)
    assert res

    a = Matrix(V, V)
    res = 2 * a
    assert isinstance(res, FormSum)
    assert res


def test_adjoint():
    domain_2d = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    f_2d = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
    V = FunctionSpace(domain_2d, f_2d)
    a = Matrix(V, V)

    adj = adjoint(a)
    res = 2 * adj
    assert isinstance(res, FormSum)
    assert res

    res = adjoint(2 * a)
    assert isinstance(res, FormSum)
    assert isinstance(res.components()[0], Adjoint)

    # Adjoint(Adjoint(.)) = Id
    assert adjoint(adj) == a


def test_action():
    domain_2d = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    f_2d = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
    V = FunctionSpace(domain_2d, f_2d)
    domain_1d = Mesh(FiniteElement("Lagrange", interval, 1, (1,), identity_pullback, H1))
    f_1d = FiniteElement("Lagrange", interval, 1, (), identity_pullback, H1)
    U = FunctionSpace(domain_1d, f_1d)

    a = Matrix(V, U)
    b = Matrix(V, U.dual())
    u = Coefficient(U)
    u_a = Argument(U, 0)
    v = Coefficient(V)
    ustar = Cofunction(U.dual())
    u_form = u_a * dx

    res = action(a, u)
    assert res
    assert len(res.arguments()) < len(a.arguments())
    assert isinstance(res, Action)

    repeat = action(res, v)
    assert repeat
    assert len(repeat.arguments()) < len(res.arguments())

    res = action(2 * a, u)
    assert isinstance(res, FormSum)
    assert isinstance(res.components()[0], Action)

    res = action(b, u_form)
    assert res
    assert len(res.arguments()) < len(b.arguments())

    with pytest.raises(TypeError):
        res = action(a, v)

    with pytest.raises(TypeError):
        res = action(a, ustar)

    b2 = Matrix(V, U.dual())
    ustar2 = Cofunction(U.dual())
    # Check Action left-distributivity with FormSum
    res = action(b, ustar + ustar2)
    assert res == Action(b, ustar) + Action(b, ustar2)
    # Check Action right-distributivity with FormSum
    res = action(b + b2, ustar)
    assert res == Action(b, ustar) + Action(b2, ustar)

    a2 = Matrix(V, U)
    u2 = Coefficient(U)
    u3 = Coefficient(U)
    # Check Action left-distributivity with Sum
    # Add 3 Coefficients to check composition of Sum works fine since u
    # + u2 + u3 => Sum(u, Sum(u2, u3))
    res = action(a, u + u2 + u3)
    assert res == Action(a, u3) + Action(a, u) + Action(a, u2)
    # Check Action right-distributivity with Sum
    res = action(a + a2, u)
    assert res == Action(a, u) + Action(a2, u)


def test_differentiation():
    domain_2d = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    f_2d = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
    V = FunctionSpace(domain_2d, f_2d)
    domain_1d = Mesh(FiniteElement("Lagrange", interval, 1, (1,), identity_pullback, H1))
    f_1d = FiniteElement("Lagrange", interval, 1, (), identity_pullback, H1)
    U = FunctionSpace(domain_1d, f_1d)

    u = Coefficient(U)
    v = Argument(U, 0)
    vstar = Argument(U.dual(), 0)

    # -- Cofunction -- #
    w = Cofunction(U.dual())
    dwdu = expand_derivatives(derivative(w, u))
    assert isinstance(dwdu, ZeroBaseForm)
    assert dwdu.arguments() == (
        Argument(w.ufl_function_space().dual(), 0),
        Argument(u.ufl_function_space(), 1),
    )
    # Check compatibility with int/float
    assert dwdu == 0

    dwdw = expand_derivatives(derivative(w, w, vstar))
    assert dwdw == vstar

    dudw = expand_derivatives(derivative(u, w))
    # du/dw is a ufl.Zero and not a ZeroBaseForm
    # as we are not differentiating a BaseForm
    assert isinstance(dudw, Zero)
    assert dudw == 0

    # -- Coargument -- #
    dvstardu = expand_derivatives(derivative(vstar, u))
    assert isinstance(dvstardu, ZeroBaseForm)
    assert dvstardu.arguments() == vstar.arguments() + (Argument(u.ufl_function_space(), 1),)
    # Check compatibility with int/float
    assert dvstardu == 0

    # -- Matrix -- #
    M = Matrix(V, U)
    dMdu = expand_derivatives(derivative(M, u))
    assert isinstance(dMdu, ZeroBaseForm)
    assert dMdu.arguments() == M.arguments() + (Argument(u.ufl_function_space(), 2),)
    # Check compatibility with int/float
    assert dMdu == 0

    # -- Action -- #
    Ac = Action(w, u)
    dAcdu = derivative(Ac, u)
    assert dAcdu == (
        action(adjoint(derivative(w, u), derivatives_expanded=True), u, derivatives_expanded=True)
        + action(w, derivative(u, u), derivatives_expanded=True)
    )

    dAcdu = expand_derivatives(dAcdu)
    # Since dw/du = 0
    assert dAcdu == Action(w, v)

    # -- Form sum -- #
    uhat = Argument(U, 1)
    what = Argument(U, 2)
    Fs = M + inner(u * uhat, v) * dx
    dFsdu = expand_derivatives(derivative(Fs, u))
    # Distribute differentiation over FormSum components
    assert dFsdu == inner(what * uhat, v) * dx


def test_zero_base_form_mult():
    domain_2d = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    f_2d = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
    V = FunctionSpace(domain_2d, f_2d)
    v = Argument(V, 0)
    Z = ZeroBaseForm((v, v))

    u = Coefficient(V)

    Zu = Z * u
    assert Zu == action(Z, u)
    assert action(Zu, u) == ZeroBaseForm(())


def test_base_form_call():
    domain_2d = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    f_2d = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
    V = FunctionSpace(domain_2d, f_2d)

    # Check duality pairing
    f = Coefficient(V)
    c = Cofunction(V.dual())
    assert c(f) == action(c, f)



"""Classes used to group scalar expressions into expressions with rank > 0."""
# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Massimiliano Leoni, 2016.

from ufl.constantvalue import Zero, as_ufl
from ufl.core.expr import Expr
from ufl.core.multiindex import FixedIndex, Index, MultiIndex, indices
from ufl.core.operator import Operator
from ufl.core.ufl_type import ufl_type
from ufl.index_combination_utils import remove_indices
from ufl.indexed import Indexed

# --- Classes representing tensors of UFL expressions ---


@ufl_type(is_shaping=True, num_ops="varying", inherit_indices_from_operand=0)
class ListTensor(Operator):
    """Wraps a list of expressions into a tensor valued expression of one higher rank."""

    __slots__ = ()

    def __new__(cls, *expressions):
        """Create a new ListTensor."""
        # All lists and tuples should already be unwrapped in
        # as_tensor
        if any(not isinstance(e, Expr) for e in expressions):
            raise ValueError("Expecting only UFL expressions in ListTensor constructor.")

        # Get properties of the first expression
        e0 = expressions[0]
        sh = e0.ufl_shape
        fi = e0.ufl_free_indices
        fid = e0.ufl_index_dimensions

        # Obviously, each subexpression must have the same shape
        if any(sh != e.ufl_shape for e in expressions[1:]):
            raise ValueError(
                "Cannot create a tensor by joining subexpressions with different shapes."
            )
        if any(fi != e.ufl_free_indices for e in expressions[1:]):
            raise ValueError(
                "Cannot create a tensor where the components have different free indices."
            )
        if any(fid != e.ufl_index_dimensions for e in expressions[1:]):
            raise ValueError(
                "Cannot create a tensor where the components have different free index dimensions."
            )

        # Simplify to Zero if possible
        if all(isinstance(e, Zero) for e in expressions):
            shape = (len(expressions),) + sh
            return Zero(shape, fi, fid)

        return Operator.__new__(cls)

    def __init__(self, *expressions):
        """Initialise."""
        Operator.__init__(self, expressions)

        # Checks
        indexset = set(self.ufl_operands[0].ufl_free_indices)
        if not all(not (indexset ^ set(e.ufl_free_indices)) for e in self.ufl_operands):
            raise ValueError(
                "Can't combine subtensor expressions with different sets of free indices."
            )

    @property
    def ufl_shape(self):
        """Get the UFL shape."""
        return (len(self.ufl_operands),) + self.ufl_operands[0].ufl_shape

    def evaluate(self, x, mapping, component, index_values, derivatives=()):
        """Evaluate."""
        if len(component) != len(self.ufl_shape):
            raise ValueError(
                "Can only evaluate scalars, expecting a component "
                "tuple of length {len(self.ufl_shape)}, not {component}."
            )
        a = self.ufl_operands[component[0]]
        component = component[1:]
        if derivatives:
            return a.evaluate(x, mapping, component, index_values, derivatives)
        else:
            return a.evaluate(x, mapping, component, index_values)

    def __getitem__(self, key):
        """Get an item."""
        origkey = key

        if isinstance(key, MultiIndex):
            key = key.indices()
        if not isinstance(key, tuple):
            key = (key,)
        k = key[0]
        if isinstance(k, (int, FixedIndex)):
            sub = self.ufl_operands[int(k)]
            return sub if len(key) == 1 else sub[key[1:]]

        return Expr.__getitem__(self, origkey)

    def __str__(self):
        """Format as a string."""

        def substring(expressions, indent):
            ind = " " * indent
            if any(isinstance(e, ListTensor) for e in expressions):
                substrings = []
                for e in expressions:
                    if isinstance(e, ListTensor):
                        substrings.append(substring(e.ufl_operands, indent + 2))
                    else:
                        substrings.append(str(e))
                s = (",\n" + ind).join(substrings)
                return "%s[\n%s%s\n%s]" % (ind, ind, s, ind)
            else:
                s = ", ".join(str(e) for e in expressions)
                return "%s[%s]" % (ind, s)

        return substring(self.ufl_operands, 0)


@ufl_type(is_shaping=True, num_ops="varying")
class ComponentTensor(Operator):
    """Maps the free indices of a scalar valued expression to tensor axes."""

    __slots__ = ("ufl_shape", "ufl_free_indices", "ufl_index_dimensions")

    def __new__(cls, expression, indices):
        """Create a new ComponentTensor."""
        # Simplify
        if isinstance(expression, Zero):
            fi, fid, sh = remove_indices(
                expression.ufl_free_indices,
                expression.ufl_index_dimensions,
                [ind.count() for ind in indices],
            )
            return Zero(sh, fi, fid)

        # Construct
        return Operator.__new__(cls)

    def __init__(self, expression, indices):
        """Initialise."""
        if not isinstance(expression, Expr):
            raise ValueError("Expecting ufl expression.")
        if expression.ufl_shape != ():
            raise ValueError("Expecting scalar valued expression.")
        if not isinstance(indices, MultiIndex):
            raise ValueError("Expecting a MultiIndex.")
        if not all(isinstance(i, Index) for i in indices):
            raise ValueError(f"Expecting sequence of Index objects, not {indices._ufl_err_str_()}.")

        Operator.__init__(self, (expression, indices))

        fi, fid, sh = remove_indices(
            expression.ufl_free_indices,
            expression.ufl_index_dimensions,
            [ind.count() for ind in indices],
        )
        self.ufl_free_indices = fi
        self.ufl_index_dimensions = fid
        self.ufl_shape = sh

    def _ufl_expr_reconstruct_(self, expressions, indices):
        """Reconstruct."""
        # Special case for simplification as_tensor(A[ii], ii) -> A
        if isinstance(expressions, Indexed):
            A, ii = expressions.ufl_operands
            if indices == ii:
                return A
        return Operator._ufl_expr_reconstruct_(self, expressions, indices)

    def indices(self):
        """Get indices."""
        return self.ufl_operands[1]

    def evaluate(self, x, mapping, component, index_values):
        """Evaluate."""
        indices = self.ufl_operands[1]
        a = self.ufl_operands[0]

        if len(indices) != len(component):
            raise ValueError("Expecting a component matching the indices tuple.")

        # Map component to indices
        for i, c in zip(indices, component):
            index_values.push(i, c)

        a = a.evaluate(x, mapping, (), index_values)

        for _ in component:
            index_values.pop()

        return a

    def __str__(self):
        """Format as a string."""
        return "{ A | A_{%s} = %s }" % (self.ufl_operands[1], self.ufl_operands[0])


# --- User-level functions to wrap expressions in the correct way ---


def numpy2nestedlists(arr):
    """Convert Numpy array to a nested list."""
    from numpy import ndarray

    if not isinstance(arr, ndarray):
        return arr
    return [numpy2nestedlists(arr[k]) for k in range(arr.shape[0])]


def _as_list_tensor(expressions):
    """Convert to a list tensor."""
    if isinstance(expressions, (list, tuple)):
        expressions = [_as_list_tensor(e) for e in expressions]
        return ListTensor(*expressions)
    else:
        return as_ufl(expressions)


def from_numpy_to_lists(expressions):
    """Convert Numpy array to lists."""
    try:
        import numpy as np

        if isinstance(expressions, np.ndarray):
            if expressions.shape == ():
                # Unwrap scalar ndarray
                return expressions.item()
            else:
                expressions = numpy2nestedlists(expressions)
    except Exception:
        pass
    return expressions


def as_tensor(expressions, indices=None):
    """Make a tensor valued expression.

    This works in two different ways, by using indices or lists.

    1) Returns :math:`A` such that :math:`A` [*indices*] = *expressions*.
    If *indices* are provided, *expressions* must be a scalar
    valued expression with all the provided indices among
    its free indices. This operator will then map each of these
    indices to a tensor axis, thereby making a tensor valued
    expression from a scalar valued expression with free indices.

    2) Returns :math:`A` such that :math:`A[k,...]` = *expressions*[k].
    If no indices are provided, *expressions* must be a list
    or tuple of expressions. The expressions can also consist
    of recursively nested lists to build higher rank tensors.
    """
    if indices is None:
        # Allow as_tensor(as_tensor(A)) and as_vector(as_vector(v)) in user code
        if isinstance(expressions, Expr):
            return expressions

        # Support numpy array, but avoid importing numpy if not needed
        if not isinstance(expressions, (list, tuple)):
            expressions = from_numpy_to_lists(expressions)

        # Sanity check
        if not isinstance(expressions, (list, tuple, Expr)):
            raise ValueError("Expecting nested list or tuple.")

        # Recursive conversion from nested lists to nested ListTensor
        # objects
        return _as_list_tensor(expressions)
    else:
        # Make sure we have a tuple of indices
        if isinstance(indices, list):
            indices = tuple(indices)
        elif not isinstance(indices, tuple):
            indices = (indices,)

        # Special case for as_tensor(expr, ii) with ii = ()
        if indices == ():
            return expressions

        indices = MultiIndex(indices)

        # Special case for simplification as_tensor(A[ii], ii) -> A
        if isinstance(expressions, Indexed):
            A, ii = expressions.ufl_operands
            if indices.indices() == ii.indices():
                return A

        # Make a tensor from given scalar expression with free indices
        return ComponentTensor(expressions, indices)


def as_matrix(expressions, indices=None):
    """As *as_tensor()*, but limited to rank 2 tensors."""
    if indices is None:
        # Allow as_matrix(as_matrix(A)) in user code
        if isinstance(expressions, Expr):
            if len(expressions.ufl_shape) != 2:
                raise ValueError("Expecting rank 2 tensor.")
            return expressions

        # To avoid importing numpy unneeded, it's quite slow...
        if not isinstance(expressions, (list, tuple)):
            expressions = from_numpy_to_lists(expressions)

        # Check for expected list structure
        if not isinstance(expressions, (list, tuple)):
            raise ValueError("Expecting nested list or tuple of Exprs.")
        if not isinstance(expressions[0], (list, tuple)):
            raise ValueError("Expecting nested list or tuple of Exprs.")
    else:
        if len(indices) != 2:
            raise ValueError("Expecting exactly two indices.")

    return as_tensor(expressions, indices)


def as_vector(expressions, index=None):
    """As ``as_tensor()``, but limited to rank 1 tensors."""
    if index is None:
        # Allow as_vector(as_vector(v)) in user code
        if isinstance(expressions, Expr):
            if len(expressions.ufl_shape) != 1:
                raise ValueError("Expecting rank 1 tensor.")
            return expressions

        # To avoid importing numpy unneeded, it's quite slow...
        if not isinstance(expressions, (list, tuple)):
            expressions = from_numpy_to_lists(expressions)

        # Check for expected list structure
        if not isinstance(expressions, (list, tuple)):
            raise ValueError("Expecting nested list or tuple of Exprs.")
    else:
        if not isinstance(index, Index):
            raise ValueError("Expecting a single Index object.")
        index = (index,)

    return as_tensor(expressions, index)


def as_scalar(expression):
    """As scalar.

    Given a scalar or tensor valued expression A, returns either of the tuples::
      (a,b) = (A, ())
      (a,b) = (A[indices], indices)
    such that a is always a scalar valued expression.
    """
    ii = indices(len(expression.ufl_shape))
    if ii:
        expression = expression[ii]
    return expression, ii


def as_scalars(*expressions):
    """As scalars.

    Given multiple scalar or tensor valued expressions A, returns either of the tuples::
      (a,b) = (A, ())
      (a,b) = ([A[0][indices], ..., A[-1][indices]], indices)
    such that a is always a list of scalar valued expressions.
    """
    ii = indices(len(expressions[0].ufl_shape))
    if ii:
        expressions = [expression[ii] for expression in expressions]
    return expressions, ii


def unit_list(i, n):
    """Create a list of zeros where the ith entry is 1."""
    return [(1 if i == j else 0) for j in range(n)]


def unit_list2(i, j, n):
    """Creage a two dimensional list of zeros where the (i,j)th entry is 1."""
    return [[(1 if (i == i0 and j == j0) else 0) for j0 in range(n)] for i0 in range(n)]


def unit_vector(i, d):
    """A constant unit vector in direction *i* with dimension *d*."""
    return as_vector(unit_list(i, d))


def unit_vectors(d):
    """A tuple of constant unit vectors in all directions with dimension *d*."""
    return tuple(unit_vector(i, d) for i in range(d))


def unit_matrix(i, j, d):
    """A constant unit matrix in direction *i*,*j* with dimension *d*."""
    return as_matrix(unit_list2(i, j, d))


def unit_matrices(d):
    """A tuple of constant unit matrices in all directions with dimension *d*."""
    return tuple(unit_matrix(i, j, d) for i in range(d) for j in range(d))


def unit_indexed_tensor(shape, component):
    """Unit indexed tensor."""
    from ufl.constantvalue import Identity
    from ufl.operators import outer  # a bit of circular dependency issue here

    r = len(shape)
    if r == 0:
        return 0, ()
    jj = indices(r)
    es = []
    for i in range(r):
        s = shape[i]
        c = component[i]
        j = jj[i]
        e = Identity(s)[c, j]
        es.append(e)
    E = es[0]
    for e in es[1:]:
        E = outer(E, e)
    return E, jj


def unwrap_list_tensor(lt):
    """Unwrap a list tensor."""
    components = []
    sh = lt.ufl_shape
    subs = lt.ufl_operands
    if len(sh) == 1:
        for s in range(sh[0]):
            components.append(((s,), subs[s]))
    else:
        for s, sub in enumerate(subs):
            for c, v in unwrap_list_tensor(sub):
                components.append(((s,) + c, v))
    return components



#
# Author: Marie Rognes
# Modified by: Martin Sandve Alnes
# Date: 2009-02-12
#
from ufl import (
    FacetNormal,
    FunctionSpace,
    Mesh,
    TestFunctions,
    TrialFunctions,
    div,
    dot,
    ds,
    dx,
    tetrahedron,
)
from ufl.finiteelement import FiniteElement, MixedElement
from ufl.pullback import contravariant_piola, identity_pullback
from ufl.sobolevspace import H1, HDiv

cell = tetrahedron
RT = FiniteElement("Raviart-Thomas", cell, 1, (3,), contravariant_piola, HDiv)
DG = FiniteElement("DG", cell, 0, (), identity_pullback, H1)
MX = MixedElement([RT, DG])
domain = Mesh(FiniteElement("Lagrange", cell, 1, (3,), identity_pullback, H1))
space = FunctionSpace(domain, MX)

(u, p) = TrialFunctions(space)
(v, q) = TestFunctions(space)

n = FacetNormal(domain)

a0 = (dot(u, v) + div(u) * q + div(v) * p) * dx
a1 = (dot(u, v) + div(u) * q + div(v) * p) * dx - p * dot(v, n) * ds

forms = [a0, a1]



"""This module defines classes representing constant values."""

# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Anders Logg, 2011.
# Modified by Massimiliano Leoni, 2016.

import numbers
from math import atan2

import ufl

# --- Helper functions imported here for compatibility---
from ufl.checks import is_python_scalar, is_true_ufl_scalar, is_ufl_scalar  # noqa: F401
from ufl.core.expr import Expr
from ufl.core.multiindex import FixedIndex, Index
from ufl.core.terminal import Terminal
from ufl.core.ufl_type import ufl_type

# Precision for float formatting
precision = None


def format_float(x):
    """Format float value based on global UFL precision."""
    if precision:
        return "{:.{prec}}".format(float(x), prec=precision)
    else:
        return "{}".format(float(x))


# --- Base classes for constant types ---


@ufl_type(is_abstract=True)
class ConstantValue(Terminal):
    """Constant value."""

    __slots__ = ()

    def __init__(self):
        """Initialise."""
        Terminal.__init__(self)

    def is_cellwise_constant(self):
        """Return whether this expression is spatially constant over each cell."""
        return True

    def ufl_domains(self):
        """Return tuple of domains related to this terminal object."""
        return ()


# TODO: Add geometric dimension/domain and Argument dependencies to Zero?
@ufl_type(is_literal=True)
class Zero(ConstantValue):
    """Representation of a zero valued expression.

    Class for representing zero tensors of different shapes.
    """

    __slots__ = ("ufl_shape", "ufl_free_indices", "ufl_index_dimensions")

    _cache = {}

    def __getnewargs__(self):
        """Get new args."""
        return (self.ufl_shape, self.ufl_free_indices, self.ufl_index_dimensions)

    def __new__(cls, shape=(), free_indices=(), index_dimensions=None):
        """Create new Zero."""
        if free_indices:
            self = ConstantValue.__new__(cls)
        else:
            self = Zero._cache.get(shape)
            if self is not None:
                return self
            self = ConstantValue.__new__(cls)
            Zero._cache[shape] = self
        self._init(shape, free_indices, index_dimensions)
        return self

    def __init__(self, shape=(), free_indices=(), index_dimensions=None):
        """Initialise."""
        pass

    def _init(self, shape=(), free_indices=(), index_dimensions=None):
        """Initialise."""
        ConstantValue.__init__(self)

        if not all(isinstance(i, int) for i in shape):
            raise ValueError("Expecting tuple of int.")
        if not isinstance(free_indices, tuple):
            raise ValueError(f"Expecting tuple for free_indices, not {free_indices}.")

        self.ufl_shape = shape
        if not free_indices:
            self.ufl_free_indices = ()
            self.ufl_index_dimensions = ()
        elif all(isinstance(i, Index) for i in free_indices):  # Handle old input format
            if not isinstance(index_dimensions, dict) and all(
                isinstance(i, Index) for i in index_dimensions.keys()
            ):
                raise ValueError(f"Expecting tuple of index dimensions, not {index_dimensions}")
            self.ufl_free_indices = tuple(sorted(i.count() for i in free_indices))
            self.ufl_index_dimensions = tuple(
                d for i, d in sorted(index_dimensions.items(), key=lambda x: x[0].count())
            )
        else:  # Handle new input format
            if not all(isinstance(i, int) for i in free_indices):
                raise ValueError(f"Expecting tuple of integer free index ids, not {free_indices}")
            if not isinstance(index_dimensions, tuple) and all(
                isinstance(i, int) for i in index_dimensions
            ):
                raise ValueError(
                    f"Expecting tuple of integer index dimensions, not {index_dimensions}"
                )

            # Assuming sorted now to avoid this cost, enable for debugging:
            # if sorted(free_indices) != list(free_indices):
            #    raise ValueError("Expecting sorted input. Remove this check later for efficiency.")

            self.ufl_free_indices = free_indices
            self.ufl_index_dimensions = index_dimensions

    def evaluate(self, x, mapping, component, index_values):
        """Evaluate."""
        return 0.0

    def __str__(self):
        """Format as a string."""
        if self.ufl_shape == () and self.ufl_free_indices == ():
            return "0"
        if self.ufl_free_indices == ():
            return "0 (shape %s)" % (self.ufl_shape,)
        if self.ufl_shape == ():
            return "0 (index labels %s)" % (self.ufl_free_indices,)
        return "0 (shape %s, index labels %s)" % (self.ufl_shape, self.ufl_free_indices)

    def __repr__(self):
        """Representation."""
        r = "Zero(%s, %s, %s)" % (
            repr(self.ufl_shape),
            repr(self.ufl_free_indices),
            repr(self.ufl_index_dimensions),
        )
        return r

    def __eq__(self, other):
        """Check equalty."""
        if isinstance(other, Zero):
            if self is other:
                return True
            return (
                self.ufl_shape == other.ufl_shape
                and self.ufl_free_indices == other.ufl_free_indices
                and self.ufl_index_dimensions == other.ufl_index_dimensions
            )
        elif isinstance(other, (int, float)):
            return other == 0
        else:
            return False

    def __neg__(self):
        """Negate."""
        return self

    def __abs__(self):
        """Absolute value."""
        return self

    def __bool__(self):
        """Convert to a bool."""
        return False

    __nonzero__ = __bool__

    def __float__(self):
        """Convert to a float."""
        return 0.0

    def __int__(self):
        """Convert to an int."""
        return 0

    def __complex__(self):
        """Convert to a complex number."""
        return 0 + 0j


def zero(*shape):
    """UFL literal constant: Return a zero tensor with the given shape."""
    if len(shape) == 1 and isinstance(shape[0], tuple):
        return Zero(shape[0])
    else:
        return Zero(shape)


# --- Scalar value types ---


@ufl_type(is_abstract=True, is_scalar=True)
class ScalarValue(ConstantValue):
    """A constant scalar value."""

    __slots__ = ("_value",)

    def __init__(self, value):
        """Initialise."""
        ConstantValue.__init__(self)
        self._value = value

    def value(self):
        """Get the value."""
        return self._value

    def evaluate(self, x, mapping, component, index_values):
        """Evaluate."""
        return self._value

    def __eq__(self, other):
        """Check equalty.

        This is implemented to allow comparison with python scalars.

        Note that this will make IntValue(1) != FloatValue(1.0),
        but ufl-python comparisons like
            IntValue(1) == 1.0
            FloatValue(1.0) == 1
        can still succeed. These will however not have the same
        hash value and therefore not collide in a dict.
        """
        if isinstance(other, self._ufl_class_):
            return self._value == other._value
        elif isinstance(other, (int, float)):
            # FIXME: Disallow this, require explicit 'expr ==
            # IntValue(3)' instead to avoid ambiguities!
            return other == self._value
        else:
            return False

    def __str__(self):
        """Format as a string."""
        return str(self._value)

    def __float__(self):
        """Convert to a float."""
        return float(self._value)

    def __int__(self):
        """Convert to an int."""
        return int(self._value)

    def __complex__(self):
        """Convert to a complex number."""
        return complex(self._value)

    def __neg__(self):
        """Negate."""
        return type(self)(-self._value)

    def __abs__(self):
        """Absolute value."""
        return type(self)(abs(self._value))

    def real(self):
        """Real part."""
        return self._value.real

    def imag(self):
        """Imaginary part."""
        return self._value.imag


@ufl_type(wraps_type=complex, is_literal=True)
class ComplexValue(ScalarValue):
    """Representation of a constant, complex scalar."""

    __slots__ = ()

    def __getnewargs__(self):
        """Get new args."""
        return (self._value,)

    def __new__(cls, value):
        """Create a new ComplexValue."""
        if value.imag == 0:
            if value.real == 0:
                return Zero()
            else:
                return FloatValue(value.real)
        else:
            return ConstantValue.__new__(cls)

    def __init__(self, value):
        """Initialise."""
        ScalarValue.__init__(self, complex(value))

    def modulus(self):
        """Get the modulus."""
        return abs(self.value())

    def argument(self):
        """Get the argument."""
        return atan2(self.value().imag, self.value().real)

    def __repr__(self):
        """Representation."""
        r = "%s(%s)" % (type(self).__name__, repr(self._value))
        return r

    def __float__(self):
        """Convert to a float."""
        raise TypeError("ComplexValues cannot be cast to float")

    def __int__(self):
        """Convert to an int."""
        raise TypeError("ComplexValues cannot be cast to int")


@ufl_type(is_abstract=True, is_scalar=True)
class RealValue(ScalarValue):
    """Abstract class used to differentiate real values from complex ones."""

    __slots__ = ()


@ufl_type(wraps_type=float, is_literal=True)
class FloatValue(RealValue):
    """Representation of a constant scalar floating point value."""

    __slots__ = ()

    def __getnewargs__(self):
        """Get new args."""
        return (self._value,)

    def __new__(cls, value):
        """Create a new FloatValue."""
        if value == 0.0:
            # Always represent zero with Zero
            return Zero()
        return ConstantValue.__new__(cls)

    def __init__(self, value):
        """Initialise."""
        super(FloatValue, self).__init__(float(value))

    def __repr__(self):
        """Representation."""
        r = "%s(%s)" % (type(self).__name__, format_float(self._value))
        return r


@ufl_type(wraps_type=int, is_literal=True)
class IntValue(RealValue):
    """Representation of a constant scalar integer value."""

    __slots__ = ()

    _cache = {}

    def __getnewargs__(self):
        """Get new args."""
        return (self._value,)

    def __new__(cls, value):
        """Create a new IntValue."""
        if value == 0:
            # Always represent zero with Zero
            return Zero()
        elif abs(value) < 100:
            # Small numbers are cached to reduce memory usage
            # (fly-weight pattern)
            self = IntValue._cache.get(value)
            if self is not None:
                return self
            self = RealValue.__new__(cls)
            IntValue._cache[value] = self
        else:
            self = RealValue.__new__(cls)
        self._init(value)
        return self

    def _init(self, value):
        """Initialise."""
        super(IntValue, self).__init__(int(value))

    def __init__(self, value):
        """Initialise."""
        pass

    def __repr__(self):
        """Representation."""
        r = "%s(%s)" % (type(self).__name__, repr(self._value))
        return r


# --- Identity matrix ---


@ufl_type()
class Identity(ConstantValue):
    """Representation of an identity matrix."""

    __slots__ = ("_dim", "ufl_shape")

    def __init__(self, dim):
        """Initialise."""
        ConstantValue.__init__(self)
        self._dim = dim
        self.ufl_shape = (dim, dim)

    def evaluate(self, x, mapping, component, index_values):
        """Evaluate."""
        a, b = component
        return 1 if a == b else 0

    def __getitem__(self, key):
        """Get an item."""
        if len(key) != 2:
            raise ValueError("Size mismatch for Identity.")
        if all(isinstance(k, (int, FixedIndex)) for k in key):
            return IntValue(1) if (int(key[0]) == int(key[1])) else Zero()
        return Expr.__getitem__(self, key)

    def __str__(self):
        """Format as a string."""
        return "I"

    def __repr__(self):
        """Representation."""
        r = "Identity(%d)" % self._dim
        return r

    def __eq__(self, other):
        """Check equalty."""
        return isinstance(other, Identity) and self._dim == other._dim


# --- Permutation symbol ---


@ufl_type()
class PermutationSymbol(ConstantValue):
    """Representation of a permutation symbol.

    This is also known as the Levi-Civita symbol, antisymmetric symbol,
    or alternating symbol.
    """

    __slots__ = ("ufl_shape", "_dim")

    def __init__(self, dim):
        """Initialise."""
        ConstantValue.__init__(self)
        self._dim = dim
        self.ufl_shape = (dim,) * dim

    def evaluate(self, x, mapping, component, index_values):
        """Evaluate."""
        return self.__eps(component)

    def __getitem__(self, key):
        """Get an item."""
        if len(key) != self._dim:
            raise ValueError("Size mismatch for PermutationSymbol.")
        if all(isinstance(k, (int, FixedIndex)) for k in key):
            return self.__eps(key)
        return Expr.__getitem__(self, key)

    def __str__(self):
        """Format as a string."""
        return "eps"

    def __repr__(self):
        """Representation."""
        r = "PermutationSymbol(%d)" % self._dim
        return r

    def __eq__(self, other):
        """Check equalty."""
        return isinstance(other, PermutationSymbol) and self._dim == other._dim

    def __eps(self, x):
        """Get eps.

        This function body is taken from
        http://www.mathkb.com/Uwe/Forum.aspx/math/29865/N-integer-Levi-Civita
        """
        result = IntValue(1)
        for i, x1 in enumerate(x):
            for j in range(i + 1, len(x)):
                x2 = x[j]
                if x1 > x2:
                    result = -result
                elif x1 == x2:
                    return Zero()
        return result


def as_ufl(expression):
    """Converts expression to an Expr if possible."""
    if isinstance(expression, (Expr, ufl.BaseForm)):
        return expression
    elif isinstance(expression, numbers.Integral):
        return IntValue(expression)
    elif isinstance(expression, numbers.Real):
        return FloatValue(expression)
    elif isinstance(expression, numbers.Complex):
        return ComplexValue(expression)
    else:
        raise ValueError(
            f"Invalid type conversion: {expression} can not be converted to any UFL type."
        )



#
# Implemented by imitation of
#  http://code.google.com/p/debiosee/wiki/DemosOptiocFlowHornSchunck
# but not tested so this could contain errors!
#
from ufl import (
    Coefficient,
    Constant,
    FunctionSpace,
    Mesh,
    derivative,
    dot,
    dx,
    grad,
    inner,
    triangle,
)
from ufl.finiteelement import FiniteElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1

# Finite element spaces for scalar and vector fields
cell = triangle
S = FiniteElement("Lagrange", cell, 1, (), identity_pullback, H1)
V = FiniteElement("Lagrange", cell, 1, (2,), identity_pullback, H1)
domain = Mesh(FiniteElement("Lagrange", cell, 1, (2,), identity_pullback, H1))
S_space = FunctionSpace(domain, S)
V_space = FunctionSpace(domain, V)

# Optical flow function
u = Coefficient(V_space)

# Previous image brightness
I0 = Coefficient(S_space)

# Current image brightness
I1 = Coefficient(S_space)

# Regularization parameter
lamda = Constant(domain)

# Coefficiental to minimize
M = (dot(u, grad(I1)) + (I1 - I0)) ** 2 * dx + lamda * inner(grad(u), grad(u)) * dx

# Derived linear system
L = derivative(M, u)
a = derivative(L, u)
L = -L
